import argparse
import json
import unittest
from unittest.mock import Mock, patch

import tufup
import tufup.repo.cli
import tufup.utils
from tests import TempDirTestCase


class ParserTests(unittest.TestCase):
    def test_get_parser(self):
        parser = tufup.repo.cli.get_parser()
        for cmd in [
            '--version',
            'init',
            'init --debug',
            'targets add 1.0 c:\\my_bundle_dir c:\\private_keys',
            'targets -d add 1.0 c:\\my_bundle_dir c:\\private_keys',
            'targets -d add -r 1.0 c:\\my_bundle_dir c:\\private_keys',
            'targets -d add -s 1.0 c:\\my_bundle_dir c:\\private_keys',
            'targets remove-latest c:\\private_keys',
            'keys my-key-name',  # todo: doesn't do anything... use subcommand?
            'keys my-key-name -c -e',
            'keys my-key-name add root c:\\private_keys d:\\more_private_keys',
            'keys my-key-name -c -e add root c:\\private_keys',
            'keys my-key-name replace old-key-name c:\\private_keys',
            'keys my-key-name -c -e replace old-key-name c:\\private_keys',
            'sign root c:\\private_keys d:\\other_private_keys',
            'sign root c:\\private_keys -e',
            'sign root c:\\private_keys -e 100',
        ]:
            with self.subTest(msg=cmd):
                args = cmd.split()
                options = parser.parse_args(args)
                expected_func_name = '_cmd_' + args[0]
                if args[0] in ['targets', 'keys']:
                    self.assertTrue(hasattr(options, 'subcommand'))
                if args[:2] == ['targets', 'add']:
                    self.assertTrue(hasattr(options, 'meta'))
                if args[0] == '--version':
                    self.assertTrue(options.version)
                else:
                    self.assertFalse(options.version)
                    self.assertEqual(expected_func_name, options.func.__name__)

    def test_get_parser_incomplete_commands(self):
        parser = tufup.repo.cli.get_parser()
        for cmd in [
            'targets',
            'targets add',
            'targets remove-latest',
            'keys',
            'keys my-key-name add',
            'keys my-key-name replace',
            'sign',
        ]:
            with self.subTest(msg=cmd):
                args = cmd.split()
                with self.assertRaises(SystemExit):
                    parser.parse_args(args)

    def test_get_parser_meta_json(self):
        parser = tufup.repo.cli.get_parser()
        args = 'targets add 1.0 c:\\my_bundle_dir c:\\private_keys'.split()
        with self.subTest(msg='no metadata'):
            self.assertIsNone(parser.parse_args(args).meta)
        json_object = '{"changes": ["line 1", "line2"]}'
        with self.subTest(msg='valid json object'):
            options = parser.parse_args(args + ['-m', json_object])
            self.assertEqual(json.loads(json_object), options.meta)
        with self.subTest(msg='invalid json'):
            with self.assertRaises(SystemExit):
                parser.parse_args(args + ['-m', '{'])
        with self.subTest(msg='valid json but not an object'):
            with self.assertRaises(SystemExit):
                parser.parse_args(args + ['-m', '["item 1", "item 2"]'])


class CommandTests(TempDirTestCase):
    def setUp(self) -> None:
        super().setUp()
        role_names = ['root', 'targets', 'snapshot', 'timestamp']
        self.config = dict(
            app_name='my-app',
            repo_dir=self.temp_dir_path / 'repo',
            keys_dir=self.temp_dir_path / 'keys',
            key_map={name: [name] for name in role_names},
            encrypted_keys=[],
            expiration_days={name: 1 for name in role_names},
        )
        mock_return_config = Mock(return_value=self.config)
        mock_keys = Mock()
        mock_keys.create_key_pair = Mock()
        mock_repo = Mock(keys=mock_keys, **self.config)
        mock_repo.save_config = Mock()
        mock_repo.initialize = Mock()
        mock_repo.add_bundle = Mock()
        mock_repo.remove_latest_bundle = Mock()
        mock_repo.add_key = Mock()
        mock_repo.replace_key = Mock()
        mock_repo.refresh_expiration_date = Mock()
        mock_repo.threshold_sign = Mock()
        mock_repo.publish_changes = Mock()
        MockRepository = Mock(return_value=mock_repo)  # noqa
        MockRepository.load_config = mock_return_config
        MockRepository.from_config = Mock(return_value=mock_repo)
        self.mock_repo = mock_repo
        self.mock_repo_class = MockRepository

    def test__cmd_init(self):
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            with patch('tufup.repo.cli.input_bool', Mock(return_value=True)):
                with patch(
                    'tufup.repo.cli._get_config_from_user',
                    self.mock_repo_class.load_config,
                ):
                    tufup.repo.cli._cmd_init(options=argparse.Namespace())
        self.mock_repo.initialize.assert_called()

    def test__cmd_keys_create(self):
        options = argparse.Namespace(
            subcommand=None, new_key_name='test', encrypted=True, create=True
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_keys(options=options)
        self.mock_repo.keys.create_key_pair.assert_called()

    def test__cmd_keys_create_and_add_key(self):
        options = argparse.Namespace(
            subcommand='add',
            create=True,
            encrypted=True,
            key_dirs=['c:\\my_private_keys'],
            new_key_name='test',
            role_name='root',
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_keys(options=options)
        self.mock_repo.keys.create_key_pair.assert_called()
        self.mock_repo.add_key.assert_called()
        self.mock_repo.publish_changes.assert_called()

    def test__cmd_keys_replace_key(self):
        options = argparse.Namespace(
            subcommand='replace',
            create=True,
            encrypted=False,
            key_dirs=['c:\\my_private_keys'],
            new_key_name='some new key to be created',
            old_key_name='some old key name',
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_keys(options=options)
        self.mock_repo.replace_key.assert_called()
        self.mock_repo.keys.create_key_pair.assert_called()
        self.mock_repo.publish_changes.assert_called()

    def test__cmd_targets_add(self):
        kwargs = dict(
            subcommand='add',
            app_version='1.0',
            bundle_dir='dummy',
            key_dirs=['c:\\my_private_keys'],
            skip_patch=True,
            required=False,
            meta=dict(),
        )
        options = argparse.Namespace(**kwargs)
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_targets(options=options)
        self.mock_repo.add_bundle.assert_called_with(
            new_version=kwargs['app_version'],
            new_bundle_dir=kwargs['bundle_dir'],
            skip_patch=kwargs['skip_patch'],
            required=kwargs['required'],
            custom_metadata=kwargs['meta'],
        )
        self.mock_repo.publish_changes.assert_called_with(
            private_key_dirs=kwargs['key_dirs']
        )

    def test__cmd_targets_remove_latest(self):
        kwargs = dict(subcommand='remove-latest', key_dirs=['c:\\my_private_keys'])
        options = argparse.Namespace(**kwargs)
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_targets(options=options)
        self.mock_repo.remove_latest_bundle.assert_called()

    def test__cmd_sign_threshold(self):
        role_name = 'root'
        key_dirs = ['c:\\my_private_keys']
        options = argparse.Namespace(
            role_name=role_name, key_dirs=key_dirs, expiration_days=None
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_sign(options=options)
        self.mock_repo.threshold_sign.assert_called_with(
            role_name=role_name, private_key_dirs=key_dirs
        )

    def test__cmd_sign_expired(self):
        role_name = 'root'
        key_dirs = ['c:\\my_private_keys']
        options = argparse.Namespace(
            role_name=role_name,
            key_dirs=key_dirs,
            expiration_days='default',  # i.e. specify -e without a value
        )
        with patch('tufup.repo.cli.Repository', self.mock_repo_class):
            tufup.repo.cli._cmd_sign(options=options)
        self.mock_repo.refresh_expiration_date.assert_called_with(
            role_name=role_name, days=self.config['expiration_days'][role_name]
        )
        self.mock_repo.publish_changes.assert_called_with(private_key_dirs=key_dirs)

    def test__get_config_from_user_no_kwargs(self):
        default = ''
        yes = 'y'
        no = 'n'
        user_inputs = iter(
            [
                'my-app',  # app name
                'my_app.__version__',  # app version
                'my_module.MyBinaryDiff',  # binary_diff
                'repo/dir',  # repo dir
                'keys/dir',  # keys dir
                yes,  # keep default root key name
                yes,  # add another root key name
                'root-2',  # key name
                no,  # add another root key name
                default,  # encrypt 'root' key
                yes,  # encrypt 'root-2' key
                '365',  # expiration days
                '2',  # signature threshold
                default,  # keep default targets key name
                default,  # add another targets key name
                default,  # encrypt 'targets' key
                '30',  # expiration days
                '1',  # signature threshold
                default,  # keep default snapshot key name
                default,  # add another snapshot key name
                default,  # encrypt 'snapshot' key
                '7',  # expiration days
                '1',  # signature threshold
                default,  # keep default timestamp key name
                default,  # add another timestamp key name
                default,  # encrypt 'timestamp' key
                '1',  # expiration days
                '1',  # signature threshold
            ]
        )
        with patch('builtins.input', lambda *_, **__: next(user_inputs)):
            config_kwargs = tufup.repo.cli._get_config_from_user()
        self.assertTrue(config_kwargs)

    def test__get_config_from_user_with_kwargs(self):
        role_names = ['root', 'targets', 'snapshot', 'timestamp']
        original_kwargs = dict(
            app_name='my-app',
            app_version_attr='my_app.__version__',
            binary_diff='my_module.MyBinaryDiff',
            repo_dir='repo/dir',
            keys_dir='keys/dir',
            key_map={name: [name] for name in role_names},
            encrypted_keys=['root'],
            expiration_days={name: 1 for name in role_names},
            thresholds={name: 1 for name in role_names},
        )
        default = ''
        with patch('builtins.input', Mock(return_value=default)):
            config_kwargs = tufup.repo.cli._get_config_from_user(**original_kwargs)
        self.assertEqual(config_kwargs, original_kwargs)



import pathlib
from unittest import TestCase

from tests import _create_cwd_change_generator, TempDirTestCase


class CreateToggleGeneratorTests(TestCase):
    def test__create_toggle_cwd_generator(self):
        original_cwd = pathlib.Path.cwd()
        toggle_generator = _create_cwd_change_generator()
        # switch to temporary dir
        temp_dir = next(toggle_generator)
        self.assertNotEqual(original_cwd, temp_dir)
        self.assertNotEqual(original_cwd, pathlib.Path.cwd())
        self.assertTrue(pathlib.Path(temp_dir).exists())
        # switch back to original dir
        try:
            next(toggle_generator)
        except StopIteration:
            pass
        self.assertEqual(original_cwd, pathlib.Path.cwd())
        self.assertFalse(pathlib.Path(temp_dir).exists())


class TempDirTestCaseTests(TempDirTestCase):
    def setUp(self) -> None:
        self.original_cwd = pathlib.Path.cwd()
        super().setUp()

    def test_temporary_cwd(self):
        current_cwd = pathlib.Path.cwd()
        self.assertNotIn('tests', str(current_cwd))
        self.assertNotEqual(self.original_cwd, current_cwd)
        self.assertEqual(self.temp_dir_path.resolve(), current_cwd.resolve())



import sys

from tufup import main

# __main__.py allows us to call the package using `python -m tufup`
# https://docs.python.org/3/library/__main__.html#main-py-in-python-packages
try:
    main()  # return code 0
except Exception as e:
    sys.exit(f'error: {e}')  # return code 1



from abc import ABC, abstractmethod
import gzip
import hashlib
import logging
import pathlib
import re
from typing import Dict, get_type_hints, Optional, TypedDict, Union

import bsdiff4
from packaging.version import Version, InvalidVersion

logger = logging.getLogger(__name__)

KEY_REQUIRED = 'required'  # unlikely to be identical to user-specified key
SUFFIX_ARCHIVE = '.tar.gz'
SUFFIX_PATCH = '.patch'


class CustomMetadataDict(TypedDict):
    """
    explicitly separate custom metadata into user-specified metadata and metadata
    used by tufup internally
    """

    user: dict
    tufup: dict


def _immutable(value):
    """
    Make value immutable, recursively, so the result is hashable.

    Applies to (nested) dict, list, set, and bytearray [1] mutable sequence types.
    Everything else is passed through unaltered, so the more exotic mutable types are
    not supported.

    [1]: https://peps.python.org/pep-3137/
    """
    # recursive cases
    if isinstance(value, dict):
        return tuple((k, _immutable(v)) for k, v in value.items())
    elif isinstance(value, list):
        return tuple(_immutable(v) for v in value)
    elif isinstance(value, set):
        return frozenset(_immutable(v) for v in value)
    elif isinstance(value, bytearray):
        return bytes(value)
    # base case
    return value


class TargetMeta(object):
    filename_pattern = '{name}-{version}{suffix}'
    filename_regex = re.compile(
        r'^(?P<name>[\w-]+)-(?P<version>.+)(?P<suffix>\.tar\.gz|\.patch)$'
    )

    def __init__(
        self,
        target_path: Union[None, str, pathlib.Path] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
        is_archive: Optional[bool] = True,
        custom: Optional[CustomMetadataDict] = None,
    ):
        """
        Initialize either with target_path, or with name, version, archive.

        BEWARE: whitespace is not allowed in the filename,
        nor in the `name` or `version` arguments
        """
        super().__init__()
        if target_path is None:
            target_path = TargetMeta.compose_filename(
                name=name, version=version, is_archive=is_archive
            )
        self.target_path_str = str(target_path)  # keep the original for reference
        self.path = pathlib.Path(target_path)
        if ' ' in self.filename:
            logger.critical(
                f'invalid filename "{self.filename}": whitespace not allowed'
            )
        self._custom = custom or dict(user=dict(), tufup=dict())

    @property
    def custom(self) -> dict:
        """returns user-specified custom metadata dict"""
        return self._get_custom_metadata('user')

    @property
    def custom_internal(self) -> dict:
        """returns tufup-internal custom metadata dict"""
        return self._get_custom_metadata('tufup')

    def _get_custom_metadata(self, _key: str) -> dict:
        """
        get custom metadata in a backward-compatible manner (older versions did not
        distinguish between user-specified and internal metadata)
        """
        if isinstance(self._custom, dict):
            # check dict keys for backward compatibility
            if get_type_hints(CustomMetadataDict).keys() != self._custom.keys():
                return self._custom
            return self._custom.get(_key)
        return dict()

    def __str__(self):
        return str(self.target_path_str)

    def __repr__(self):
        return f'{self.__class__.__name__}(target_path="{self.target_path_str}")'

    def __hash__(self):
        """
        This makes the object hashable, so it can be used as dict key or set
        member.

        https://docs.python.org/3/glossary.html#term-hashable

        """
        return hash(_immutable(self.__dict__))

    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return vars(self) == vars(other)

    def __lt__(self, other):
        """
        This makes the object sortable, based on the *version* property,
        without having to specify an explicit sorting key. Note this
        disregards app name, platform, and suffixes.
        """
        if type(other) is not type(self):
            return NotImplemented
        return self.version < other.version

    @property
    def filename(self):
        return self.path.name

    @property
    def name(self) -> Optional[str]:
        """The app name."""
        match_dict = self.parse_filename(self.filename)
        return match_dict.get('name')

    @property
    def version(self) -> Optional[Version]:
        match_dict = self.parse_filename(self.filename)
        try:
            version = Version(match_dict.get('version', ''))
        except InvalidVersion:
            version = None
            logger.critical(f'No valid version in filename: {self.filename}')
        return version

    @property
    def suffix(self) -> Optional[str]:
        """Returns the filename suffix, either '.tar.gz', '.patch', or None."""
        match_dict = self.parse_filename(self.filename)
        return match_dict.get('suffix')

    @property
    def is_archive(self) -> bool:
        return self.suffix == SUFFIX_ARCHIVE

    @property
    def is_patch(self) -> bool:
        return self.suffix == SUFFIX_PATCH

    @property
    def is_other(self) -> bool:
        return self.suffix not in [SUFFIX_ARCHIVE, SUFFIX_PATCH]

    @classmethod
    def parse_filename(cls, filename: str) -> dict:
        """
        Parse a filename to extract app name, version, and suffix.

        We do not impose any versioning requirements yet, such as defined in
        packaging.version.VERSION_PATTERN.
        """
        match = cls.filename_regex.search(string=filename)
        return match.groupdict() if match else {}

    @classmethod
    def compose_filename(cls, name: str, version: str, is_archive: bool):
        suffix = SUFFIX_ARCHIVE if is_archive else SUFFIX_PATCH
        return cls.filename_pattern.format(name=name, version=version, suffix=suffix)


class BinaryDiff(ABC):
    """
    BinaryDiff represents an interface for overriding the binary diff/patch functions
    """

    @staticmethod
    @abstractmethod
    def diff(*, src_bytes: bytes, dst_bytes: bytes) -> bytes:
        """Create patch as the binary difference between source and destination data"""
        pass

    @staticmethod
    @abstractmethod
    def patch(*, src_bytes: bytes, patch_bytes: bytes) -> bytes:
        """Apply binary patch to source data to recover destination data"""
        pass


class DefaultBinaryDiff(BinaryDiff):
    """The default implementation of binary differencing and patching functions"""

    diff = bsdiff4.diff
    patch = bsdiff4.patch


class Patcher(object):
    DEFAULT_HASH_ALGORITHM = 'sha256'

    @staticmethod
    def _get_tar_size_and_hash(
        tar_content: Optional[bytes] = None, algorithm: str = DEFAULT_HASH_ALGORITHM
    ) -> dict:
        """
        Determines the size and hash of the specified data.

        Note we could also use tuf.api.metadata.TargetFile for this, but that we'll
        keep this part independent from python-tuf, for clarity and flexibility.
        """
        hash_obj = getattr(hashlib, algorithm)()
        hash_obj.update(tar_content)
        # hexdigest returns digest as string
        return dict(
            tar_size=len(tar_content),
            tar_hash=hash_obj.hexdigest(),
            tar_hash_algorithm=algorithm,
        )

    @classmethod
    def _verify_tar_size_and_hash(cls, tar_content: bytes, expected: dict):
        """
        Verifies that size and hash of data match the expected values.

        Raises an exception if this is not the case.
        """
        result = cls._get_tar_size_and_hash(
            tar_content=tar_content, algorithm=expected['tar_hash_algorithm']
        )
        for key in ['tar_size', 'tar_hash']:
            if result[key] != expected[key]:
                raise Exception(f'verification failed: {key} mismatch')

    @classmethod
    def diff_and_hash(
        cls,
        src_path: pathlib.Path,
        dst_path: pathlib.Path,
        patch_path: pathlib.Path,
        binary_diff: Optional[type[BinaryDiff]] = None,
    ) -> dict:
        """
        Creates a patch file from the binary difference between source and destination
        .tar archives. The source and destination files are expected to be
        gzip-compressed tar archives (.tar.gz).

        The binary differencing method can be customized by implementing a `BinaryDiff`
        subclass and passing this in via the `binary_diff` argument.

        Returns a dict with size and hash of the *uncompressed* destination archive.
        """
        binary_diff = binary_diff or DefaultBinaryDiff
        with gzip.open(src_path, mode='rb') as src_file:
            with gzip.open(dst_path, mode='rb') as dst_file:
                dst_tar_content = dst_file.read()
                patch_path.write_bytes(
                    binary_diff.diff(
                        src_bytes=src_file.read(), dst_bytes=dst_tar_content
                    )
                )
        return cls._get_tar_size_and_hash(tar_content=dst_tar_content)

    @classmethod
    def patch_and_verify(
        cls,
        src_path: pathlib.Path,
        dst_path: pathlib.Path,
        patch_targets: Dict[TargetMeta, pathlib.Path],
        binary_diff: Optional[type[BinaryDiff]] = None,
    ) -> None:
        """
        Applies one or more binary patch files to a source file in order to
        reconstruct a destination file.

        Source file and destination file are gzip-compressed tar archives, but the
        patches are applied to the *uncompressed* tar archives. The reason is that
        small changes in uncompressed data can cause (very) large differences in
        gzip compressed data, leading to excessively large patch files (see #69).

        The integrity of the patched .tar archive is verified using expected length
        and hash (from custom tuf metadata), similar to python-tuf's download
        verification. If the patched archive fails this check, the destination file
        is not written.

        The binary patching method can be customized by implementing a `BinaryDiff`
        subclass and passing this in via the `binary_diff` argument.
        """
        binary_diff = binary_diff or DefaultBinaryDiff
        if not patch_targets:
            raise ValueError('no patch targets')
        # decompress .tar data from source .tar.gz file
        with gzip.open(src_path, mode='rb') as src_file:
            tar_bytes = src_file.read()
        # apply cumulative patches (sorted by version, in ascending order)
        for patch_meta, patch_path in sorted(patch_targets.items()):
            logger.info(f'applying patch: {patch_meta.name}')
            tar_bytes = binary_diff.patch(
                src_bytes=tar_bytes, patch_bytes=patch_path.read_bytes()
            )
        # verify integrity of the final result (raises exception on failure)
        cls._verify_tar_size_and_hash(
            tar_content=tar_bytes,
            expected=patch_meta.custom_internal,  # noqa
        )
        # compress .tar data into destination .tar.gz file
        with gzip.open(dst_path, mode='wb') as dst_file:
            dst_file.write(tar_bytes)



import copy
import tarfile
import unittest
from datetime import date, datetime, timedelta
import json
import logging
import pathlib
from tempfile import TemporaryDirectory
from time import sleep
from unittest.mock import Mock, patch

from securesystemslib.interface import (
    generate_and_write_unencrypted_ed25519_keypair,
    generate_and_write_ed25519_keypair,
    import_ed25519_publickey_from_file,
)
from tuf.api.metadata import (
    Metadata,
    Role,
    Root,
    Targets,
    Snapshot,
    Timestamp,
    TargetFile,
    TOP_LEVEL_ROLE_NAMES,
)

from tests import TempDirTestCase, TEST_REPO_DIR
from tufup.common import DefaultBinaryDiff, KEY_REQUIRED, TargetMeta
import tufup.repo  # for patching
from tufup.repo import (
    Base,
    get_binary_diff_class,
    in_,
    Keys,
    make_gztar_archive,
    Repository,
    Roles,
    SUFFIX_PUB,
    SUFFIX_PATCH,
)
from tufup.utils.platform_specific import ON_WINDOWS

mock_input = Mock(return_value='')

DUMMY_SSL_KEY = {
    'keytype': 'ed25519',
    'scheme': 'ed25519',
    'keyid': '22f7c6046e29cfb0205a1c07941a5a57da39a6859b844f8c347f622a57ff82c8',
    'keyid_hash_algorithms': ['sha256', 'sha512'],
    'keyval': {
        'public': '93032b5804ba40a725145171193782bdfa30038584715546aea3228ea8018e46'
    },
}
DUMMY_ROOT = Root(
    version=1,
    spec_version='1.0',
    expires=datetime.now() + timedelta(days=1),
    keys=dict(),
    roles={
        role_name: Role(keyids=[], threshold=1) for role_name in TOP_LEVEL_ROLE_NAMES
    },
    consistent_snapshot=False,
)
DUMMY_EXPIRATION_DAYS = dict(root=1000, targets=100, snapshot=10, timestamp=1)
DUMMY_PRIVATE_KEY_PATHS = dict(
    (role_name, [pathlib.Path('dummy', role_name)])
    for role_name in TOP_LEVEL_ROLE_NAMES
)
DUMMY_KEY_MAP = dict(
    root=['root_one', 'root_two'],
    targets=['targets'],
    snapshot=['snapshot'],
    timestamp=['timestamp'],
)
DUMMY_THRESHOLDS = dict(root=4, targets=3, snapshot=2, timestamp=1)


class ModuleTests(TempDirTestCase):
    def test_in_(self):
        self.assertIsInstance(in_(days=1), datetime)

    def test_make_gztar_archive(self):
        app_name = 'test'
        version = '1.2.3'
        # prepare
        sub_dir = self.temp_dir_path / 'sub'
        sub_file = sub_dir / 'sub.txt'
        root_file = self.temp_dir_path / 'root.txt'
        sub_dir.mkdir()
        sub_file.touch()
        root_file.touch()
        # test
        mock_input_no = Mock(return_value='n')
        for exists in [False, True]:
            with patch('builtins.input', mock_input_no):
                archive = make_gztar_archive(
                    src_dir=self.temp_dir_path,
                    dst_dir=self.temp_dir_path,
                    app_name=app_name,
                    version=version,
                    tar_format=tarfile.USTAR_FORMAT,
                )
            self.assertIsInstance(archive, TargetMeta)
            self.assertEqual(exists, mock_input_no.called)
            self.assertTrue(archive.path.exists())
            self.assertTrue(app_name in str(archive.path))
            self.assertTrue(version in str(archive.path))

    def test_get_binary_diff_class(self):
        cases = [(None, None), ('tufup.common.DefaultBinaryDiff', DefaultBinaryDiff)]
        for name, expected in cases:
            with self.subTest(msg=name):
                self.assertIs(
                    get_binary_diff_class(fully_qualified_name=name), expected
                )


class BaseTests(TempDirTestCase):
    def test_init(self):
        with patch('builtins.input', mock_input):
            # dir exists
            base = Base(dir_path=self.temp_dir_path)
            self.assertTrue(base.dir_path.exists())
            # dir does not exist yet
            base = Base(dir_path=self.temp_dir_path / 'new')
            self.assertTrue(base.dir_path.exists())


class KeysTests(TempDirTestCase):
    def test_init_no_key_files(self):
        # no public key files exist yet
        keys = Keys(dir_path=self.temp_dir_path)
        for role_name in TOP_LEVEL_ROLE_NAMES:
            self.assertFalse(getattr(keys, role_name))

    def test_init_and_import_all_public_keys(self):
        # create some key files
        for role_name in TOP_LEVEL_ROLE_NAMES:
            private_key_filename = Keys.filename_pattern.format(key_name=role_name)
            file_path = self.temp_dir_path / private_key_filename
            generate_and_write_unencrypted_ed25519_keypair(filepath=str(file_path))
        # test
        keys = Keys(dir_path=self.temp_dir_path)
        for role_name in TOP_LEVEL_ROLE_NAMES:
            self.assertIsInstance(getattr(keys, role_name)[0], dict)

    def test_init_and_import_all_public_keys_with_key_map(self):
        # create a single key-pair
        key_name = 'single'
        private_key_filename = Keys.filename_pattern.format(key_name=key_name)
        file_path = self.temp_dir_path / private_key_filename
        generate_and_write_unencrypted_ed25519_keypair(filepath=str(file_path))
        key_map = {role_name: [key_name] for role_name in TOP_LEVEL_ROLE_NAMES}
        # test
        keys = Keys(dir_path=self.temp_dir_path, key_map=key_map)
        for role_name in TOP_LEVEL_ROLE_NAMES:
            self.assertIsInstance(getattr(keys, role_name)[0], dict)
            # all keys should be equal
            self.assertEqual(keys.root, getattr(keys, role_name))

    def test_import_all_public_keys(self):
        # note: import_all_public_keys is also (implicitly) tested via __init__
        keys = Keys(dir_path=self.temp_dir_path)
        # create some key files
        for role_name in TOP_LEVEL_ROLE_NAMES:
            private_key_filename = Keys.filename_pattern.format(key_name=role_name)
            file_path = self.temp_dir_path / private_key_filename
            generate_and_write_unencrypted_ed25519_keypair(filepath=str(file_path))
        # test
        for role_name in TOP_LEVEL_ROLE_NAMES:
            self.assertFalse(getattr(keys, role_name))
        keys.import_all_public_keys()
        for role_name in TOP_LEVEL_ROLE_NAMES:
            self.assertIsInstance(getattr(keys, role_name)[0], dict)

    def test_import_public_key(self):
        # create dummy key with name differing from role name
        key_name = 'test'
        role_name = 'targets'
        private_key_filename = Keys.filename_pattern.format(key_name=key_name)
        file_path = self.temp_dir_path / private_key_filename
        generate_and_write_unencrypted_ed25519_keypair(filepath=str(file_path))
        # test
        keys = Keys(dir_path=self.temp_dir_path)
        keys.import_public_key(role_name=role_name, key_name=key_name)
        self.assertTrue(keys.targets)

    def test_create(self):
        with patch('getpass.getpass', mock_input):
            keys = Keys(dir_path=self.temp_dir_path, encrypted=['root'])
            keys.create()
            # key pair files should now exist
            filenames = [item.name for item in keys.dir_path.iterdir()]
            for role_name in TOP_LEVEL_ROLE_NAMES:
                private_key_filename = Keys.filename_pattern.format(key_name=role_name)
                public_key_filename = private_key_filename + SUFFIX_PUB
                self.assertIn(private_key_filename, filenames)
                self.assertIn(public_key_filename, filenames)
            # and the public keys should have been imported
            self.assertTrue(all(getattr(keys, n) for n in TOP_LEVEL_ROLE_NAMES))

    def test_create_with_key_map(self):
        # prepare
        keys = Keys(dir_path=self.temp_dir_path, key_map=DUMMY_KEY_MAP)
        expected_key_names = [
            key_name for key_names in DUMMY_KEY_MAP.values() for key_name in key_names
        ]
        # test
        keys.create()
        # we should now have five key-pairs (one for each, but two for root)
        filenames = [item.name for item in keys.dir_path.iterdir()]
        self.assertEqual(2 * len(expected_key_names), len(filenames))
        for key_name in expected_key_names:
            self.assertIn(key_name, filenames)
            self.assertIn(key_name + SUFFIX_PUB, filenames)
        # and the public keys should have been imported
        for role_name, key_names in DUMMY_KEY_MAP.items():
            self.assertEqual(len(key_names), len(getattr(keys, role_name)))

    def test_create_key_pair(self):
        public_key_path = Keys.create_key_pair(
            private_key_path=self.temp_dir_path / 'key_name', encrypted=False
        )
        self.assertTrue(public_key_path.exists())

    def test_create_key_pair_do_not_overwrite(self):
        # create dummy key pair
        key_name = 'dummy'
        private_key_filename = Keys.filename_pattern.format(key_name=key_name)
        private_key_path = self.temp_dir_path / private_key_filename
        generate_and_write_unencrypted_ed25519_keypair(filepath=str(private_key_path))
        original_private_key = private_key_path.read_bytes()
        with patch('builtins.input', Mock(return_value='n')):
            Keys.create_key_pair(private_key_path=private_key_path, encrypted=False)
        self.assertEqual(original_private_key, private_key_path.read_bytes())

    def test_public(self):
        keys = Keys(dir_path=self.temp_dir_path)
        # test empty
        self.assertFalse(keys.public())
        # set a dummy key value
        keys.root = [DUMMY_SSL_KEY]
        # test
        self.assertIn(DUMMY_SSL_KEY['keyid'], keys.public().keys())

    def test_roles(self):
        keys = Keys(dir_path=self.temp_dir_path)
        # test empty
        self.assertSetEqual(set(TOP_LEVEL_ROLE_NAMES), set(keys.roles().keys()))
        # set a dummy key value
        keys.root = [DUMMY_SSL_KEY]
        # test
        self.assertIn('root', keys.roles().keys())

    def test_roles_thresholds(self):
        # prepare
        keys = Keys(dir_path=self.temp_dir_path, thresholds=DUMMY_THRESHOLDS)
        for role_name in TOP_LEVEL_ROLE_NAMES:
            setattr(keys, role_name, [DUMMY_SSL_KEY])
        # test
        roles = keys.roles()
        for key, value in roles.items():
            self.assertEqual(DUMMY_THRESHOLDS[key], value.threshold)

    def test_find_private_key(self):
        # create dummy private key files in separate folders
        key_names = [
            ('online', [Snapshot.type, Timestamp.type]),
            ('offline/subdir', [Root.type, Targets.type]),  # subdir tests recursion
        ]
        for dir_name, role_names in key_names:
            dir_path = self.temp_dir_path / dir_name
            dir_path.mkdir(parents=True)
            for role_name in role_names:
                filename = Keys.filename_pattern.format(key_name=role_name)
                (dir_path / filename).touch()
        # test
        key_dirs = list(self.temp_dir_path.iterdir())  # ['online', 'offline']
        for role_name in TOP_LEVEL_ROLE_NAMES:
            key_path = Keys.find_private_key(key_name=role_name, key_dirs=key_dirs)
            self.assertTrue(key_path)
            self.assertIn(role_name, str(key_path))
            self.assertTrue(key_path.exists())
        self.assertIsNone(Keys.find_private_key(key_name='missing', key_dirs=key_dirs))


class RolesTests(TempDirTestCase):
    def test_init(self):
        self.assertTrue(Roles(dir_path=self.temp_dir_path))

    def test_init_import_roles(self):
        def mock_from_file(filename, *args, **kwargs):
            file_path = pathlib.Path(filename)
            assert file_path.exists()
            return file_path.stem

        # create dummy metadata files
        for role_name in TOP_LEVEL_ROLE_NAMES:
            filenames = [f'{role_name}.json']
            if role_name == 'root':
                filenames.extend(f'{v}.{role_name}.json' for v in [1, 2, 3])
            for filename in filenames:
                (self.temp_dir_path / filename).touch()
        # test
        with patch.object(tufup.repo.Metadata, 'from_file', mock_from_file):
            roles = Roles(dir_path=self.temp_dir_path)
            for role_name in TOP_LEVEL_ROLE_NAMES:
                self.assertEqual(role_name, getattr(roles, role_name))

    def test_initialize_empty(self):
        # prepare
        mock_keys = Mock()
        mock_keys.public = Mock()
        mock_keys.roles = Mock(
            return_value={name: None for name in TOP_LEVEL_ROLE_NAMES}
        )
        roles = Roles(dir_path=self.temp_dir_path)
        # test
        roles.initialize(keys=mock_keys)
        self.assertTrue(
            all(isinstance(getattr(roles, n), Metadata) for n in TOP_LEVEL_ROLE_NAMES)
        )
        # files do not exist yet, because the roles still need to be populated
        self.assertFalse(any(roles.dir_path.iterdir()))

    def test_initialize_existing_root(self):
        # prepare
        mock_keys = Mock()
        mock_keys.public = Mock()
        mock_keys.roles = Mock(return_value={n: None for n in TOP_LEVEL_ROLE_NAMES})
        roles = Roles(dir_path=self.temp_dir_path)
        # set existing root role
        mock_root_role = Mock()
        roles.root = mock_root_role
        # test
        roles.initialize(keys=mock_keys)
        # ensure the existing role has not been replaced
        self.assertEqual(mock_root_role, roles.root)

    def test_add_or_update_target(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        roles.targets = Mock(signed=Mock(targets=dict()))
        # test (path must exist)
        filename = 'my_app.tar.gz'
        local_target_path = self.temp_dir_path / filename
        with self.assertRaises(FileNotFoundError):
            roles.add_or_update_target(local_path=local_target_path)
        # test (path segments)
        local_target_path.write_bytes(b'some bytes')
        cases = [
            (None, filename),
            ([], filename),  # update
            (['a', 'b'], 'a/b/' + filename),
            (['a', 'b'], 'a/b/' + filename),  # update with segments
        ]
        for segments, expected_url_path in cases:
            with self.subTest(msg=segments):
                roles.add_or_update_target(
                    local_path=local_target_path, url_path_segments=segments
                )
                self.assertIsInstance(
                    roles.targets.signed.targets[expected_url_path], TargetFile
                )
        # ensure update did not create new items
        self.assertEqual(2, len(roles.targets.signed.targets))
        # test (custom)
        custom = dict(something='whatever')
        roles.add_or_update_target(local_path=local_target_path, custom=custom)
        self.assertEqual(custom, roles.targets.signed.targets[filename].custom)

    def test_remove_target(self):
        # prepare
        filename = 'my_app-1.0.tar.gz'
        dirname = 'subdir'
        url_path = f'{dirname}/{filename}'
        subdir = self.temp_dir_path / dirname
        subdir.mkdir()
        local_target_path = subdir / filename
        local_target_path.touch()
        roles = Roles(dir_path=self.temp_dir_path)
        roles.targets = Mock(signed=Mock(targets={url_path: Mock()}))
        original_targets_signed = copy.deepcopy(roles.targets.signed)
        # test
        self.assertTrue(local_target_path.exists())
        roles.remove_target(local_path=local_target_path)
        self.assertNotIn(filename, roles.targets.signed.targets)
        self.assertFalse(local_target_path.exists())
        self.assertNotEqual(original_targets_signed, roles.targets.signed)
        # cannot remove non-existent target
        self.assertFalse(roles.remove_target(local_path=local_target_path))

    def test_add_public_key(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        roles.root = Mock(signed=Mock(roles=dict(), add_key=Mock()))
        public_key_path = self.temp_dir_path / 'targets_key.pub'
        public_key_path.write_text(json.dumps(DUMMY_SSL_KEY))
        # test
        role_name = 'targets'
        roles.add_public_key(role_name=role_name, public_key_path=public_key_path)
        self.assertTrue(roles.root.signed.add_key.called)

    def test_set_signature_threshold(self):
        # prepare
        role_name = 'targets'
        threshold = 2
        roles = Roles(dir_path=self.temp_dir_path)
        roles.root = Mock(signed=Mock(roles={role_name: Mock(threshold=1)}))
        # test
        roles.set_signature_threshold(role_name=role_name, threshold=threshold)
        self.assertEqual(threshold, roles.root.signed.roles[role_name].threshold)

    def test_sign_role(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        roles.root = Metadata(signed=DUMMY_ROOT, signatures=dict())
        role_name = 'root'
        signature_count = 2
        password = 'mock-password'
        with patch('getpass.getpass', Mock(return_value=password)):
            for index in range(signature_count):
                private_key_path = self.temp_dir_path / f'{index}.{role_name}'
                # create key pair
                if index == 0:
                    generate_and_write_unencrypted_ed25519_keypair(
                        filepath=str(private_key_path)
                    )
                else:
                    generate_and_write_ed25519_keypair(
                        password=password, filepath=str(private_key_path)
                    )
                # test
                roles.sign_role(role_name=role_name, private_key_path=private_key_path)
        self.assertEqual(signature_count, len(roles.root.signatures))

    def test_file_path(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        # test
        self.assertEqual(
            self.temp_dir_path / '1.root.json',
            roles.file_path(role_name='root', version=1),
        )
        self.assertEqual(
            self.temp_dir_path / 'root.json',
            roles.file_path(role_name='root'),
        )
        self.assertEqual(
            self.temp_dir_path / 'targets.json',
            roles.file_path(role_name='targets', version=1),
        )

    def test_file_exists(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        (self.temp_dir_path / '1.root.json').touch()
        # test
        self.assertTrue(roles.file_exists(role_name='root'))
        self.assertFalse(roles.file_exists(role_name='targets'))

    def test_persist_role(self):
        # prepare
        roles = Roles(dir_path=self.temp_dir_path)
        roles.root = Metadata(signed=DUMMY_ROOT, signatures=dict())
        expected_filenames = [f'{DUMMY_ROOT.version}.root.json', 'root.json']
        # test
        roles.persist_role(role_name='root')
        for filename in expected_filenames:
            with self.subTest(msg=filename):
                self.assertTrue((self.temp_dir_path / filename).exists())

    def test_get_latest_archive(self):
        roles = Roles(dir_path=TEST_REPO_DIR / 'metadata')
        expected_filename = 'example_app-4.0a0.tar.gz'
        latest_archive = roles.get_latest_archive()
        self.assertEqual(expected_filename, latest_archive.path.name)

    def test_get_latest_archive_empty(self):
        roles = Roles(dir_path=self.temp_dir_path)
        self.assertIsNone(roles.get_latest_archive())


class RepositoryTests(TempDirTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        logging.basicConfig(level=logging.DEBUG)

    def test_defaults(self):
        self.assertTrue(Repository(app_name='test'))

    def test_init_paths(self):
        repo_dir_name = 'repo'
        keys_dir_name = 'keystore'
        # absolute paths (could also use resolve on relative path...)
        repo_dir_abs = self.temp_dir_path / repo_dir_name
        keys_dir_abs = self.temp_dir_path / keys_dir_name
        cases = [
            ('string', repo_dir_name, keys_dir_name),
            ('relative', pathlib.Path(repo_dir_name), pathlib.Path(keys_dir_name)),
            ('absolute', repo_dir_abs, keys_dir_abs),
        ]
        for message, repo_dir, keys_dir in cases:
            with self.subTest(msg=message):
                repo = Repository(app_name='test', repo_dir=repo_dir, keys_dir=keys_dir)
                # internally we should always have the absolute paths
                self.assertTrue(repo.repo_dir.is_absolute())
                self.assertTrue(repo.keys_dir.is_absolute())
                # compare dirs
                # resolve is necessary for github actions, see:
                # https://github.com/actions/runner-images/issues/712#issuecomment-1163036706
                self.assertEqual(repo_dir_abs.resolve(), repo.repo_dir.resolve())
                self.assertEqual(keys_dir_abs.resolve(), repo.keys_dir.resolve())

    def test_config_dict(self):
        app_name = 'test'
        repo = Repository(app_name=app_name)
        expected_config_dict = {
            'app_name': app_name,
            'app_version_attr': None,
            'encrypted_keys': None,
            'expiration_days': None,
            'key_map': None,
            'keys_dir': None,
            'repo_dir': None,
            'thresholds': None,
            'binary_diff': None,
        }
        self.assertEqual(set(expected_config_dict), set(repo.config_dict))

    def test_app_version(self):
        # for convenience we use the notosotuf version attribute here,
        # but normally this would point to an external app version e.g.
        # 'my_app.__version__'
        app_version_attr = 'tufup.__version__'
        repo = Repository(app_name='test', app_version_attr=app_version_attr)
        self.assertEqual(str(tufup.__version__), repo.app_version)

    def test_get_config_file_path(self):
        self.assertTrue(Repository.get_config_file_path())

    def test_save_config(self):
        # prepare
        repo = Repository(app_name='test')
        # test
        repo.save_config()
        self.assertTrue(repo.get_config_file_path().exists())
        config_file_text = repo.get_config_file_path().read_text()
        print(config_file_text)  # for convenience
        # paths saved to config file are relative to current working
        # directory (cwd) if possible (otherwise absolute paths are saved)
        config_dict = json.loads(config_file_text)
        for key in ['repo_dir', 'keys_dir']:
            with self.subTest(msg=key):
                # note Path.is_relative_to() is introduced in python 3.9
                self.assertFalse(pathlib.Path(config_dict[key]).is_absolute())

    @unittest.skipUnless(condition=ON_WINDOWS, reason='windows only')
    def test_save_config_windows_paths(self):
        # prepare
        kwargs = dict(repo_dir='foo\\repo', keys_dir='bar\\keys')
        repo = Repository(app_name='test', **kwargs)
        # test
        repo.save_config()
        self.assertTrue(repo.get_config_file_path().exists())
        config_text = repo.get_config_file_path().read_text()
        print(config_text)
        config = json.loads(repo.get_config_file_path().read_text())
        for key in kwargs.keys():
            with self.subTest(msg=key):
                self.assertEqual(kwargs[key].replace('\\', '/'), config[key])

    def test_save_config_binary_diff(self):
        cases = [
            (None, None),
            (DefaultBinaryDiff, 'tufup.common.DefaultBinaryDiff'),
        ]
        for binary_diff, expected in cases:
            with self.subTest(msg=binary_diff):
                # prepare
                repo = Repository(app_name='test', binary_diff=binary_diff)
                # test
                repo.save_config()
                self.assertTrue(repo.get_config_file_path().exists())
                config_text = repo.get_config_file_path().read_text()
                print(config_text)
                config = json.loads(repo.get_config_file_path().read_text())
                self.assertEqual(expected, config.get('binary_diff'))

    def test_load_config(self):
        # file does not exist
        self.assertEqual(dict(), Repository.load_config())
        # file exists but invalid
        Repository.get_config_file_path().touch()
        # test
        self.assertEqual(dict(), Repository.load_config())

    @unittest.skipIf(condition=ON_WINDOWS, reason='posix only')
    def test_load_config_windows_paths(self):
        # prepare (mix windows paths and posix paths for convenience)
        mock_config = dict(repo_dir='foo\\repo', keys_dir='/tmp/bar/keys')
        config_path = Repository.get_config_file_path()
        config_path.write_text(json.dumps(mock_config))
        print(config_path.read_text())
        # test
        config = Repository.load_config()
        for key in mock_config.keys():
            with self.subTest(msg=key):
                self.assertEqual(mock_config[key].replace('\\', '/'), config[key])

    def test_from_config(self):
        temp_dir = self.temp_dir_path.resolve()
        repo_dir_abs = temp_dir / 'repo'
        keys_dir_abs = temp_dir / 'keystore'
        cases = [
            ('absolute paths', repo_dir_abs, keys_dir_abs),
            (
                'relative paths',
                repo_dir_abs.relative_to(temp_dir),
                keys_dir_abs.relative_to(temp_dir),
            ),
        ]
        for message, repo_dir, keys_dir in cases:
            with self.subTest(msg=message):
                # prepare
                config_data = dict(
                    app_name='test',
                    app_version_attr='my_app.__version__',
                    repo_dir=repo_dir,
                    keys_dir=keys_dir,
                    key_map=dict(),
                    encrypted_keys=[],
                    expiration_days=dict(),
                    thresholds=dict(),
                )
                Repository.get_config_file_path().write_text(
                    json.dumps(config_data, default=str)
                )
                # test
                with patch.object(Repository, '_load_keys_and_roles') as mmock_load:
                    repo = Repository.from_config()
                # internally the repo should always work with absolute paths
                # (relative paths are resolved in the class initializer)
                config_data['repo_dir'] = repo_dir_abs
                config_data['keys_dir'] = keys_dir_abs
                self.assertEqual(
                    config_data,
                    {key: getattr(repo, key) for key in config_data.keys()},
                )
                self.assertTrue(mmock_load.called)

    def test_from_config_binary_diff(self):
        cases = [
            (None, None),
            ('tufup.common.DefaultBinaryDiff', DefaultBinaryDiff),
        ]
        for binary_diff, expected in cases:
            with self.subTest(msg=binary_diff):
                # prepare
                config_data = dict(
                    app_name='test',
                    app_version_attr='my_app.__version__',
                    repo_dir='repo',
                    keys_dir='keystore',
                    key_map=dict(),
                    encrypted_keys=[],
                    expiration_days=dict(),
                    thresholds=dict(),
                    binary_diff=binary_diff,
                )
                Repository.get_config_file_path().write_text(
                    json.dumps(config_data, default=str)
                )
                # test
                with patch.object(Repository, '_load_keys_and_roles'):
                    repo = Repository.from_config()
                self.assertIs(repo.binary_diff, expected)

    def test_initialize(self):
        # prepare
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
            expiration_days=DUMMY_EXPIRATION_DAYS,
        )
        # test
        repo.initialize()
        self.assertTrue(any(repo.keys_dir.iterdir()))
        self.assertTrue(any(repo.metadata_dir.iterdir()))
        self.assertTrue(all(getattr(repo.roles, name) for name in TOP_LEVEL_ROLE_NAMES))
        self.assertEqual(
            date.today() + timedelta(days=DUMMY_EXPIRATION_DAYS['root']),
            repo.roles.root.signed.expires.date(),
        )

    def test_initialize_extra_key_dirs(self):
        # prepare
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
            expiration_days=DUMMY_EXPIRATION_DAYS,
        )
        repo.initialize()
        # move private keys to separate dir
        private_key_dir = self.temp_dir_path / 'private_keys'
        private_key_dir.mkdir()
        for path in repo.keys_dir.iterdir():
            if path.is_file() and not path.suffix:
                path.rename(target=private_key_dir / path.name)
        # remove metadata files
        for path in repo.metadata_dir.iterdir():
            if path.suffix == '.json':
                path.unlink()
        # reproduce issue #102
        with self.assertRaises(Exception) as context:
            repo.initialize()
        self.assertIn('no private keys found', str(context.exception).lower())
        # test fix
        try:
            repo.initialize(extra_key_dirs=[private_key_dir])
        except Exception as e:
            self.fail(msg=f'unexpected exception: {e}')

    def test_refresh_expiration_date(self):
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        days = 999
        repo.refresh_expiration_date(role_name='root', days=days)
        self.assertEqual(
            date.today() + timedelta(days=days),
            repo.roles.root.signed.expires.date(),
        )

    def test_replace_key(self):
        role_name = 'root'
        # prepare
        new_key_name = 'new_key'
        old_key_name = 'old-key'
        key_map = {name: [name] for name in TOP_LEVEL_ROLE_NAMES}
        key_map['root'].append(old_key_name)
        key_map['targets'] = [old_key_name]
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
            key_map=key_map,
        )
        repo.initialize()  # todo: make test independent...
        old_key_id = repo.roles.root.signed.roles['targets'].keyids[0]
        # create new key pair to replace old one
        new_private_key_path = repo.keys_dir / new_key_name
        new_public_key_path = Keys.create_key_pair(
            private_key_path=new_private_key_path, encrypted=False
        )
        new_key_id = import_ed25519_publickey_from_file(
            filepath=str(new_public_key_path)
        )['keyid']
        # test
        expected_key_count = len(key_map[role_name])
        repo.replace_key(
            old_key_name=old_key_name,
            new_public_key_path=new_public_key_path,
            new_private_key_encrypted=True,  # pretend the key is encrypted
        )
        self.assertEqual(
            expected_key_count,
            len(repo.roles.root.signed.roles[role_name].keyids),
        )
        # old key removed?
        self.assertNotIn(old_key_id, repo.roles.root.signed.roles[role_name].keyids)
        self.assertNotIn(old_key_name, repo.key_map[role_name])
        self.assertIn(old_key_name, repo.revoked_key_names)
        # new key added?
        self.assertIn(new_key_id, repo.roles.root.signed.roles[role_name].keyids)
        self.assertIn(new_key_name, repo.key_map[role_name])
        self.assertIn(new_key_name, repo.encrypted_keys)
        # no duplicates in encrypted_keys?
        self.assertEqual(1, len(repo.encrypted_keys))
        # other keys still in key map?
        self.assertEqual(expected_key_count, len(repo.key_map[role_name]))
        # keys replaced for all roles?
        self.assertIn(new_key_name, repo.key_map['targets'])

    def test_add_key(self):
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        # create new key pair to add
        new_key_name = 'new-key'
        new_private_key_path = repo.keys_dir / new_key_name
        new_public_key_path = Keys.create_key_pair(
            private_key_path=new_private_key_path, encrypted=False
        )
        new_key_id = import_ed25519_publickey_from_file(
            filepath=str(new_public_key_path)
        )['keyid']
        # test
        role_name = 'root'
        repo.add_key(
            role_name=role_name,
            public_key_path=new_public_key_path,
            encrypted=True,  # pretend the key is encrypted
        )
        # new key added?
        self.assertEqual(2, len(repo.roles.root.signed.roles[role_name].keyids))
        self.assertIn(new_key_id, repo.roles.root.signed.roles[role_name].keyids)
        self.assertIn(new_key_name, repo.key_map[role_name])
        self.assertIn(new_key_name, repo.encrypted_keys)

    def test_add_bundle(self):
        app_name = 'test'
        version = '1.0'
        # prepare
        bundle_dir = self.temp_dir_path / 'dist' / 'test_app'
        bundle_dir.mkdir(parents=True)
        bundle_file = bundle_dir / 'dummy.exe'
        bundle_file.touch()
        repo = Repository(
            app_name=app_name,
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        # test
        repo.add_bundle(
            new_version=version,
            new_bundle_dir=bundle_dir,
            custom_metadata=dict(whatever='something'),
            required=True,
        )
        self.assertTrue((repo.metadata_dir / 'targets.json').exists())
        target_name = f'{app_name}-{version}.tar.gz'
        self.assertTrue(repo.roles.targets.signed.targets[target_name].custom)
        self.assertTrue(
            repo.roles.targets.signed.targets[target_name].custom['tufup'][KEY_REQUIRED]
        )

    def test_add_bundle_no_patch(self):
        # prepare
        bundle_dir = self.temp_dir_path / 'dist' / 'test_app'
        bundle_dir.mkdir(parents=True)
        bundle_file = bundle_dir / 'dummy.exe'
        bundle_file.write_text('this is version 1')
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        repo.add_bundle(new_version='1.0', new_bundle_dir=bundle_dir)
        # test
        bundle_file.write_text('much has changed in version 2')
        repo.add_bundle(new_version='2.0', new_bundle_dir=bundle_dir, skip_patch=True)
        self.assertTrue((repo.metadata_dir / 'targets.json').exists())
        target_keys = list(repo.roles.targets.signed.targets.keys())
        self.assertEqual(2, len(target_keys))
        self.assertFalse(any(key.endswith(SUFFIX_PATCH) for key in target_keys))

    def test_remove_latest_bundle(self):
        # prepare
        bundle_dir = self.temp_dir_path / 'dist' / 'test_app'
        bundle_dir.mkdir(parents=True)
        bundle_file = bundle_dir / 'dummy.exe'
        bundle_file.touch()
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        v1 = '1.0'
        v2 = '2.0'
        repo.add_bundle(new_version=v1, new_bundle_dir=bundle_dir)
        repo.add_bundle(new_version=v2, new_bundle_dir=bundle_dir)
        # test
        repo.remove_latest_bundle()
        remaining_target_keys = list(repo.roles.targets.signed.targets.keys())
        self.assertEqual(1, len(remaining_target_keys))
        self.assertIn(v1, remaining_target_keys[0])
        remaining_target_filenames = [p.name for p in repo.targets_dir.iterdir()]
        self.assertEqual(1, len(remaining_target_filenames))
        self.assertIn(remaining_target_keys[0], remaining_target_filenames)

    def test_threshold_sign(self):
        # prepare
        role_name = 'root'
        keys_dir = self.temp_dir_path / 'keystore'
        repo = Repository(
            app_name='test',
            keys_dir=keys_dir,
            repo_dir=self.temp_dir_path / 'repo',
        )
        repo.initialize()  # todo: make test independent...
        # create dummy "revoked key" and pretend it is encrypted
        revoked_key_name = 'revoked'
        repo.keys.create_key_pair(
            private_key_path=keys_dir / revoked_key_name, encrypted=False
        )
        repo.revoked_key_names.append(revoked_key_name)
        repo.encrypted_keys.append(revoked_key_name)
        # dummy modification
        repo.roles.root.signed.spec_version = 'x'
        versioned_file_path = repo.metadata_dir / f'1.{role_name}.json'
        non_versioned_file_path = repo.metadata_dir / f'{role_name}.json'
        versioned_last_modified = versioned_file_path.stat().st_mtime_ns
        non_versioned_last_modified = non_versioned_file_path.stat().st_mtime_ns
        # test
        sleep(0.1)  # enforce different modification time
        count = repo.threshold_sign(
            role_name=role_name, private_key_dirs=[repo.keys_dir]
        )
        self.assertEqual(2, count)  # existing key and revoked key
        self.assertFalse(repo.revoked_key_names)
        self.assertFalse(repo.encrypted_keys)
        # files should have been modified
        self.assertGreater(
            versioned_file_path.stat().st_mtime_ns, versioned_last_modified
        )
        self.assertGreater(
            non_versioned_file_path.stat().st_mtime_ns, non_versioned_last_modified
        )
        # if no signatures found
        for path in repo.keys_dir.iterdir():
            path.unlink()
        with self.assertRaises(Exception):
            repo.threshold_sign(role_name=role_name, private_key_dirs=[repo.keys_dir])

    def test__load_keys_and_roles(self):
        # prepare
        keys_dir = self.temp_dir_path / 'keystore'
        repo = Repository(
            app_name='test',
            keys_dir=keys_dir,
            repo_dir=self.temp_dir_path / 'repo',
        )
        # test
        with patch('builtins.input', Mock(return_value='y')):
            repo._load_keys_and_roles(create_keys=True)
        self.assertTrue(all(getattr(repo.roles, name) for name in TOP_LEVEL_ROLE_NAMES))
        self.assertTrue(
            all((keys_dir / name).exists() for name in TOP_LEVEL_ROLE_NAMES)
        )

    def test_publish_changes_no_change(self):
        # prepare
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        # note that initialize() already calls publish_changes...
        repo.initialize()  # todo: make test independent...
        # test
        repo.publish_changes(private_key_dirs=[repo.keys_dir])
        for role_name in ['root', 'targets', 'snapshot', 'timestamp']:
            role = getattr(repo.roles, role_name)
            with self.subTest(msg=role_name):
                self.assertEqual(1, role.signed.version)
                self.assertTrue(role.signatures)

    def test_publish_changes_threshold(self):
        # prepare
        repo = Repository(
            app_name='test',
            keys_dir=self.temp_dir_path / 'keystore',
            repo_dir=self.temp_dir_path / 'repo',
        )
        # note that initialize() already calls publish_changes...
        repo.initialize()  # todo: make test independent...
        # remove the root signature
        repo.roles.root.signatures = {}
        # test
        repo.publish_changes(private_key_dirs=[repo.keys_dir])
        for role_name in ['root', 'targets', 'snapshot', 'timestamp']:
            with self.subTest(msg=role_name):
                role = getattr(repo.roles, role_name)
                self.assertEqual(repo.thresholds[role_name], len(role.signatures))

    def test_publish_changes(self):
        days = 999
        role_names = ['root', 'targets', 'snapshot', 'timestamp']
        for test_role_name in role_names.copy():
            # use a fresh temporary dir and repo
            with TemporaryDirectory() as temp_dir:
                temp_dir_path = pathlib.Path(temp_dir)
                repo = Repository(
                    app_name='test',
                    keys_dir=temp_dir_path / 'keystore',
                    repo_dir=temp_dir_path / 'repo',
                )
                # note that initialize() already calls publish_changes...
                repo.initialize()  # todo: make test independent...
                # make a change to metadata (in memory)
                repo.roles.set_expiration_date(role_name=test_role_name, days=days)
                # make a change to config
                config_change = 'dummy'
                repo.encrypted_keys.append(config_change)
                # test
                repo.publish_changes(private_key_dirs=[repo.keys_dir])
                for role_name in role_names:
                    role = getattr(repo.roles, role_name)
                    with self.subTest(msg=f'{test_role_name}: {role_name}'):
                        self.assertEqual(2, role.signed.version)
                        self.assertTrue(role.signatures)
                role_names.remove(test_role_name)
                #  re-load from file to verify change
                root = Metadata.from_file(
                    repo.roles.file_path(role_name=test_role_name)
                )
                self.assertEqual(
                    date.today() + timedelta(days=days),
                    root.signed.expires.date(),
                )
                # verify change in config
                config_from_disk = repo.load_config()
                self.assertIn(config_change, config_from_disk['encrypted_keys'])



import gzip
import hashlib
import logging
import pathlib
from typing import Hashable
from unittest import TestCase

import bsdiff4
from packaging.version import Version

from tests import TempDirTestCase
from tufup.common import _immutable, Patcher, TargetMeta  # noqa


class ImmutableTests(TestCase):
    def test_immutable(self):
        cases = [
            'a',
            b'b',
            1,
            ('a', 1),
            ['a', 1],
            {'a', 1},
            bytearray(b'b'),
            dict(a=1),
            [dict(a=[dict(c={1})], b=bytearray(b'd'))],
            dict(a=dict(b=dict(c=dict()))),
        ]
        for case in cases:
            with self.subTest(msg=case):
                self.assertIsInstance(_immutable(case), Hashable)


class TargetMetaTests(TempDirTestCase):
    def test_init_whitespace(self):
        for kwargs in [
            dict(target_path='w h i t e s p a c e-1.2.3.tar.gz'),
            dict(name='w h i t e s p a c e'),
        ]:
            with self.subTest(msg=kwargs):
                with self.assertLogs(level=logging.CRITICAL):
                    TargetMeta(**kwargs)

    def test_eq_ne(self):
        target_meta = TargetMeta()
        self.assertEqual(target_meta, TargetMeta())
        self.assertNotEqual(target_meta, TargetMeta(target_path='something'))
        # two ways to initialize a TargetMeta instance:
        self.assertEqual(
            TargetMeta(target_path='my-app-1.2.3a0.patch'),
            TargetMeta(name='my-app', version='1.2.3a0', is_archive=False),
        )

    def test_repr(self):
        # https://docs.python.org/3/reference/datamodel.html#object.__repr__
        target_meta = TargetMeta(target_path='some.file')
        # the input to eval is known and trusted, so that should be safe, right?
        self.assertEqual(eval(repr(target_meta)), target_meta)

    def test_str(self):
        # see issue #3
        for target_path in ['str_path', pathlib.Path('pathlib_path')]:
            with self.subTest(msg=target_path):
                self.assertIsInstance(
                    TargetMeta(target_path=target_path).__str__(), str
                )

    def test_hashable(self):
        obj = TargetMeta()
        try:
            obj.__hash__()
        except Exception as e:
            self.fail(f'__hash__ failed unexpectedly: {e}')
        expected_obj_hashable = (
            ('target_path_str', 'None-None.tar.gz'),
            ('path', pathlib.Path('None-None.tar.gz')),
            ('_custom', (('user', ()), ('tufup', ()))),
        )
        self.assertEqual(hash(expected_obj_hashable), obj.__hash__())
        # we can use the obj as a set member or as dict key
        self.assertEqual({obj, obj}, {obj})

    def test_sortable(self):
        version_1 = TargetMeta(target_path='my-app-win-1.0.tar.gz')
        version_2 = TargetMeta(target_path='my-app-win-2.0.tar.gz')
        info_list = [version_2, version_1]
        self.assertEqual([version_1, version_2], sorted(info_list))

    def test_filename(self):
        target_path = 'url/path/somefile'
        target_meta = TargetMeta(target_path=target_path)
        self.assertEqual('somefile', target_meta.filename)

    def test_name(self):
        target_meta = TargetMeta(target_path='url/path/my-app-1.0.tar.gz')
        self.assertEqual('my-app', target_meta.name)

    def test_version(self):
        for version_str in ['1.2.3a4', '']:
            expected = Version(version_str) if version_str else None
            with self.subTest(msg=version_str):
                target_meta = TargetMeta(target_path=f'x-{version_str}.tar.gz')
                self.assertEqual(expected, target_meta.version)

    def test_suffix(self):
        valid_suffixes = ['.tar.gz', '.patch']
        for suffix in ['', '.zip', '.tar.gz', '.patch']:
            with self.subTest(msg=suffix):
                target_meta = TargetMeta(target_path=f'my-app-1.2.3a4{suffix}')
                if suffix in valid_suffixes:
                    self.assertEqual(suffix, target_meta.suffix)
                else:
                    self.assertIsNone(target_meta.suffix)

    def test_is_archive(self):
        self.assertTrue(TargetMeta(target_path='my-app-1.0.tar.gz').is_archive)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.patch').is_archive)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.zip').is_archive)

    def test_is_patch(self):
        self.assertTrue(TargetMeta(target_path='my-app-1.0.patch').is_patch)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.tar.gz').is_patch)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.zip').is_patch)

    def test_is_other(self):
        self.assertTrue(TargetMeta(target_path='my-app-1.0.zip').is_other)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.patch').is_other)
        self.assertFalse(TargetMeta(target_path='my-app-1.0.tar.gz').is_other)

    def test_parse_filename(self):
        cases = [
            # PEP440 versions are allowed (we do not parse them here...)
            ('app-1.patch', ('app', '1', '.patch')),
            ('app-1.tar.gz', ('app', '1', '.tar.gz')),
            ('app-1.2.tar.gz', ('app', '1.2', '.tar.gz')),
            ('app-1.2.3.tar.gz', ('app', '1.2.3', '.tar.gz')),
            ('app-1a.tar.gz', ('app', '1a', '.tar.gz')),
            ('app-1b.tar.gz', ('app', '1b', '.tar.gz')),
            ('app-1rc.tar.gz', ('app', '1rc', '.tar.gz')),
            ('app-1rc0.tar.gz', ('app', '1rc0', '.tar.gz')),
            ('app-2022.0.tar.gz', ('app', '2022.0', '.tar.gz')),
            # we don't impose a specific version format at this point (that
            # is deferred to packaging.Version)
            ('app-invalidversion.tar.gz', ('app', 'invalidversion', '.tar.gz')),
            # underscores, dashes, capitals, etc.
            ('app-name---1.tar.gz', ('app-name--', '1', '.tar.gz')),
            ('CAPS-1.tar.gz', ('CAPS', '1', '.tar.gz')),
            ('un_der_scores-1.0.tar.gz', ('un_der_scores', '1.0', '.tar.gz')),
            # invalid filenames
            ('sp ac es-1.zip', ()),
            ('app-1.gz', ()),
            ('app-1.xz', ()),
            ('anything', ()),
        ]
        for filename, expected in cases:
            match_dict = TargetMeta.parse_filename(filename=filename)
            with self.subTest(msg=filename):
                self.assertEqual(expected, tuple(match_dict.values()))

    def test_compose_filename(self):
        filename = TargetMeta.compose_filename(
            name='app', version='1.0', is_archive=True
        )
        self.assertEqual('app-1.0.tar.gz', filename)

    def test_custom_metadata(self):
        user_metadata = dict(foo='bar')
        internal_metadata = dict(something=True)
        target_meta = TargetMeta(
            custom=dict(user=user_metadata, tufup=internal_metadata)
        )
        self.assertEqual(user_metadata, target_meta.custom)
        self.assertEqual(internal_metadata, target_meta.custom_internal)

    def test_custom_metadata_backward_compatibility(self):
        # older versions of tufup did not distinguish between user and internal metadata
        custom_metadata = dict(foo='bar')
        target_meta = TargetMeta(custom=custom_metadata)  # noqa
        self.assertEqual(custom_metadata, target_meta.custom)
        self.assertEqual(custom_metadata, target_meta.custom_internal)

    def test_custom_metadata_not_specified(self):
        target_meta = TargetMeta()
        self.assertIsInstance(target_meta.custom, dict)
        self.assertIsInstance(target_meta.custom_internal, dict)
        # user overrides _custom attr
        target_meta._custom = None
        self.assertIsInstance(target_meta.custom, dict)


class PatcherTests(TempDirTestCase):
    def setUp(self) -> None:
        super().setUp()
        # define dummy .tar content
        self.tar_content = {
            'v-1': b'this is the original content',
            'v-2': b'this content is somewhat different',
            'v-3': b'this content has changed again',
        }
        # create patch content
        self.patch_content = dict()
        for src, dst in [('v-1', 'v-2'), ('v-2', 'v-3')]:
            self.patch_content[dst] = bsdiff4.diff(
                src_bytes=self.tar_content[src],
                dst_bytes=self.tar_content[dst],
            )
        # create dummy files
        self.targz_paths = dict()
        for key, tar_content in self.tar_content.items():
            self.targz_paths[key] = self.temp_dir_path / f'{key}.tar.gz'
            with gzip.open(self.targz_paths[key], mode='wb') as gz_file:
                gz_file.write(tar_content)
        self.patch_paths = dict()
        for key, patch_content in self.patch_content.items():
            self.patch_paths[key] = self.temp_dir_path / f'{key}.patch'
            self.patch_paths[key].write_bytes(patch_content)
        # determine size and hash
        hash_algorithm = 'sha256'
        self.tar_fingerprints = dict()
        for key, tar_content in self.tar_content.items():
            if key == 'v-1':
                continue
            hash_obj = getattr(hashlib, hash_algorithm)()
            hash_obj.update(tar_content)
            self.tar_fingerprints[key] = dict(
                tar_size=len(tar_content),
                tar_hash=hash_obj.hexdigest(),
                tar_hash_algorithm=hash_algorithm,
            )

    def test_diff_and_hash(self):
        # todo: explicitly test the binary_diff argument?
        # prepare
        src = 'v-1'
        dst = 'v-2'
        patch_path = self.temp_dir_path / 'test.patch'
        # test
        dst_fingerprint = Patcher.diff_and_hash(
            src_path=self.targz_paths[src],
            dst_path=self.targz_paths[dst],
            patch_path=patch_path,
        )
        self.assertTrue(patch_path.exists())
        self.assertEqual(self.patch_content[dst], patch_path.read_bytes())
        self.assertEqual(self.tar_fingerprints[dst], dst_fingerprint)

    def test_patch_and_verify(self):
        # todo: explicitly test the binary_diff argument?
        # prepare
        src = 'v-1'
        dst = 'v-3'  # note we're skipping v-2
        patch_targets = dict()
        for key, patch_path in self.patch_paths.items():
            patch_meta = TargetMeta(
                target_path=patch_path, custom=self.tar_fingerprints[key]
            )
            self.assertTrue(patch_meta.is_patch)  # just to be sure
            patch_targets[patch_meta] = patch_path
        # verify that we're applying two patches cumulatively
        self.assertEqual(2, len(patch_targets))
        # test
        dst_path = self.temp_dir_path / 'reconstructed.tar.gz'
        Patcher.patch_and_verify(
            src_path=self.targz_paths[src],
            dst_path=dst_path,
            patch_targets=patch_targets,
        )
        self.assertTrue(dst_path.exists())
        # note that gzip compressed files are not reproducible by default (even when
        # using identical uncompressed data), so we must compare the uncompressed data
        with gzip.open(self.targz_paths[dst], mode='rb') as original_tar:
            with gzip.open(dst_path, mode='rb') as reconstructed_tar:
                self.assertEqual(original_tar.read(), reconstructed_tar.read())



import logging
import pathlib
import platform
import shutil
import subprocess
import sys
from tempfile import NamedTemporaryFile
from typing import List, Optional, Union

from tufup.utils import remove_path

logger = logging.getLogger(__name__)

CURRENT_PLATFORM = platform.system()
ON_WINDOWS = CURRENT_PLATFORM == 'Windows'
ON_MAC = CURRENT_PLATFORM == 'Darwin'
PLATFORM_SUPPORTED = ON_WINDOWS or ON_MAC


def install_update(
    src_dir: Union[pathlib.Path, str],
    dst_dir: Union[pathlib.Path, str],
    purge_dst_dir: bool = False,
    exclude_from_purge: Optional[List[Union[pathlib.Path, str]]] = None,
    **kwargs,
):
    r"""
    Installs update files using platform specific installation script. The
    actual installation script copies the files and folders from `src_dir` to
    `dst_dir`.

    If `purge_dst_dir` is `True`, *ALL* files and folders are deleted from
    `dst_dir` before copying.

    **DANGER**:

    ONLY use `purge_dst_dir=True` if your app is properly installed in its
    own *separate* directory, such as `%PROGRAMFILES%\MyApp`.

    DO NOT use `purge_dst_dir=True` if your app executable is running
    directly from a folder that also contains unrelated files or folders,
    such as the Desktop folder or the Downloads folder, because this
    unrelated content would be then also be deleted.

    Individual files and folders can be excluded from purge using e.g.

        exclude_from_purge=['path\\to\\file1', r'"path to\file2"', ...]

    If `purge_dst_dir` is `False`, the `exclude_from_purge` argument is
    ignored.
    """
    if ON_WINDOWS:
        _install_update = _install_update_win
    elif ON_MAC:
        _install_update = _install_update_mac
    else:
        raise RuntimeError('This platform is not supported.')
    return _install_update(
        src_dir=src_dir,
        dst_dir=dst_dir,
        purge_dst_dir=purge_dst_dir,
        exclude_from_purge=exclude_from_purge,
        **kwargs,
    )


# Note that robocopy itself also has an option to create a log file,
# viz. `/log:<filename>`, as well as a `/tee` option, but we want to log *all*
# output from the batch file, not just output from the robocopy command.
WIN_LOG_LINES = """
call :log > "{log_file_path}" 2>&1
:log
"""
WIN_ROBOCOPY_OVERWRITE = (
    '/e',  # include subdirectories, even if empty
    '/move',  # deletes files and dirs from source dir after they've been copied
    '/v',  # verbose (show what is going on)
    '/w:2',  # set retry-timeout (default is 30 seconds)
)
WIN_ROBOCOPY_PURGE = '/purge'  # delete all files and dirs in destination folder
WIN_ROBOCOPY_EXCLUDE_FROM_PURGE = '/xf'  # exclude specified paths from purge
# makes batch file delete itself when done (https://stackoverflow.com/a/20333575)
WIN_BATCH_DELETE_SELF = '(goto) 2>nul & del "%~f0"'

# _install_update_win makes sure the following variables are available for
# batch templates:
# {log_lines}, {src_dir}, {dst_dir}, {robocopy_options}, {delete_self}
WIN_BATCH_TEMPLATE = """@echo off
{log_lines}
echo Moving app files...
robocopy "{src_dir}" "{dst_dir}" {robocopy_options}
echo Done.
{delete_self}
"""
WIN_BATCH_PREFIX = 'tufup'
WIN_BATCH_SUFFIX = '.bat'


def run_bat_as_admin(file_path: Union[pathlib.Path, str]):
    """
    Request elevation for windows command interpreter (opens UAC prompt) and
    then run the specified .bat file.

    Returns True if successfully started, does not block, can continue after
    calling process exits.
    """
    from ctypes import windll

    # https://docs.microsoft.com/en-us/windows/win32/api/shellapi/nf-shellapi-shellexecutew
    result = windll.shell32.ShellExecuteW(
        None,  # handle to parent window
        'runas',  # verb
        'cmd.exe',  # file on which verb acts
        ' '.join(['/c', f'"{file_path}"']),  # parameters
        None,  # working directory (default is cwd)
        1,  # show window normally
    )
    success = result > 32
    if not success:
        logger.error(
            f'failed to run batch script as admin (ShellExecuteW returned {result})'
        )
    return success


def _install_update_win(
    src_dir: Union[pathlib.Path, str],
    dst_dir: Union[pathlib.Path, str],
    purge_dst_dir: bool,
    exclude_from_purge: List[Union[pathlib.Path, str]],
    as_admin: bool = False,
    batch_template: str = WIN_BATCH_TEMPLATE,
    batch_template_extra_kwargs: Optional[dict] = None,
    log_file_name: Optional[str] = None,
    robocopy_options_override: Optional[List[str]] = None,
    process_creation_flags=None,
    **kwargs,  # noqa
):
    """
    Create a batch script that moves files from src to dst, then run the
    script in a new console, and exit the current process.

    The script is created in a default temporary directory, and deletes
    itself when done.

    The `as_admin` options allows installation as admin (opens UAC dialog).

    The `batch_template` option allows specification of custom batch-file
    content. This may be in the form of a template string, as in the default
    `WIN_BATCH_TEMPLATE`, or it may be a ready-made string without any
    template variables. The following default template variables are
    available for use in the custom template, although their use is optional:
    {log_lines}, {src_dir}, {dst_dir}, {robocopy_options}, {delete_self}.
    Custom template variables can be used as well, in which case you'll need
    to specify `batch_template_extra_kwargs`.

    The `batch_template_extra_kwargs` options allows specification of
    *custom* template variables (in addition to the default ones, which are
    always available). It accepts a dict, with keys matching the *custom*
    template variable names specified in the `batch_template`.

    The `log_file_name` option will log the output of the install script to a
    file in the `dst_dir`.

    The `robocopy_options_override` option allows options for [robocopy][1]
    to be overridden completely. It accepts a list of option strings. This
    will cause the purge arguments to be ignored as well.

    The `process_creation_flags` option allows users to override creation flags for
    the subprocess call that runs the batch script. For example, one could specify
    `subprocess.CREATE_NO_WINDOW` to prevent a window from opening. See [2] and [3]
    for details.

    [1]: https://docs.microsoft.com/en-us/windows-server/administration/windows-commands/robocopy
    [2]: https://docs.python.org/3/library/subprocess.html#windows-constants
    [3]: https://learn.microsoft.com/en-us/windows/win32/procthread/process-creation-flags
    """
    if batch_template_extra_kwargs is None:
        batch_template_extra_kwargs = dict()
    # collect robocopy options
    if robocopy_options_override is None:
        robocopy_options = list(WIN_ROBOCOPY_OVERWRITE)
        if purge_dst_dir:
            robocopy_options.append(WIN_ROBOCOPY_PURGE)
            if exclude_from_purge:
                robocopy_options.append(WIN_ROBOCOPY_EXCLUDE_FROM_PURGE)
                robocopy_options.extend(exclude_from_purge)
    else:
        # empty list [] simply clears all options
        robocopy_options = robocopy_options_override
    options_str = ' '.join(robocopy_options)
    # handle batch file output logging
    log_lines = ''
    if log_file_name:
        log_file_path = pathlib.Path(dst_dir) / log_file_name
        log_lines = WIN_LOG_LINES.format(log_file_path=log_file_path)
        logger.info(f'logging install script output to {log_file_path}')
    # write temporary batch file (NOTE: The file is placed in the system
    # default temporary dir, but the file is not removed automatically. So,
    # either the batch file should self-delete when done, or it should be
    # deleted by some other means, because windows does not clean the temp
    # dir automatically.)
    script_content = batch_template.format(
        src_dir=src_dir,
        dst_dir=dst_dir,
        robocopy_options=options_str,
        log_lines=log_lines,
        delete_self=WIN_BATCH_DELETE_SELF,
        **batch_template_extra_kwargs,
    )
    logger.debug(f'writing windows batch script:\n{script_content}')
    with NamedTemporaryFile(
        mode='w', prefix=WIN_BATCH_PREFIX, suffix=WIN_BATCH_SUFFIX, delete=False
    ) as temp_file:
        temp_file.write(script_content)
    logger.debug(f'temporary batch script created: {temp_file.name}')
    script_path = pathlib.Path(temp_file.name).resolve()
    logger.debug(f'starting script in new console: {script_path}')
    # start the script in a separate process, non-blocking
    if as_admin:
        logger.debug('as admin')
        run_bat_as_admin(file_path=script_path)
    else:
        # by default we create a new console with window, but user can override this
        # using the process_creation_flags argument
        if process_creation_flags is None:
            process_creation_flags = subprocess.CREATE_NEW_CONSOLE
        else:
            logger.debug('using custom process creation flags')
        # we use Popen() instead of run(), because the latter blocks execution
        subprocess.Popen([script_path], creationflags=process_creation_flags)
    logger.debug('exiting')
    # exit current process
    sys.exit(0)


def _install_update_mac(
    src_dir: Union[pathlib.Path, str],
    dst_dir: Union[pathlib.Path, str],
    purge_dst_dir: bool,
    exclude_from_purge: List[Union[pathlib.Path, str]],
    symlinks: bool = False,
    **kwargs,
):
    """
    The symlinks arg is passed on to shutil.copytree()

    [1]: https://docs.python.org/3/library/shutil.html#shutil.copytree
    """
    # todo: implement as_admin and debug kwargs for mac
    logger.debug(f'Kwargs not used: {kwargs}')
    if purge_dst_dir:
        exclude_from_purge = (
            [  # enforce path objects
                pathlib.Path(item) for item in exclude_from_purge
            ]
            if exclude_from_purge
            else []
        )
        logger.debug(f'Purging content of {dst_dir}')
        for path in pathlib.Path(dst_dir).iterdir():
            if path not in exclude_from_purge:
                remove_path(path=path)
    logger.debug(f'Moving content of {src_dir} to {dst_dir}.')
    shutil.copytree(src_dir, dst_dir, dirs_exist_ok=True, symlinks=symlinks)
    # Note: the src_dir is typically a temporary directory, but we'll clear
    # it anyway just to be consistent with the windows implementation
    for path in pathlib.Path(src_dir).iterdir():
        remove_path(path=path)
    logger.debug(f'Restarting application, running {sys.executable}.')
    subprocess.Popen(sys.executable, shell=True)  # nosec
    sys.exit(0)


def _patched_resolve(path: pathlib.Path):
    """
    this is a rather crude workaround for cpython issue #82852,
    where Path.resolve() yields a relative path, on windows, if the target
    does not exist yet

    https://github.com/python/cpython/issues/82852

    todo: remove this as soon as support for python 3.9 is dropped
    """
    if ON_WINDOWS and sys.version_info[:2] < (3, 10):
        logger.warning('using patched path for cpython #82852')
        if not path.is_absolute():
            path = pathlib.Path.cwd() / path
    return path.resolve()



import filecmp
import gzip
import logging
import os
import pathlib
import shutil
from typing import Optional
import unittest
from unittest.mock import Mock, patch

import packaging.version
from requests.auth import HTTPBasicAuth
import tuf.api.exceptions
from tuf.ngclient import TargetFile

from tests import TempDirTestCase, TEST_REPO_DIR
import tufup.client
from tufup.client import AuthRequestsFetcher, Client, SUFFIX_FAILED
from tufup.common import TargetMeta

ROOT_FILENAME = 'root.json'
TARGETS_FILENAME = 'targets.json'
SNAPSHOT_FILENAME = 'snapshot.json'
TIMESTAMP_FILENAME = 'timestamp.json'
ON_GITHUB = os.getenv('GITHUB_ACTIONS')


class ClientTests(TempDirTestCase):
    def setUp(self) -> None:
        super().setUp()
        # directory where the parent application is installed (e.g.
        # %PROGRAMFILES%\MyApp or %LOCALAPPDATA%\Programs\MyApp on Windows 10)
        # https://docs.microsoft.com/en-us/windows/win32/msi/installation-context
        self.app_install_dir = self.temp_dir_path / 'programs' / 'example'
        self.app_install_dir.mkdir(parents=True)
        # directories must be created by parent application
        self.metadata_dir = self.temp_dir_path / 'metadata'
        self.target_dir = self.temp_dir_path / 'targets'
        self.metadata_dir.mkdir()
        self.target_dir.mkdir()
        # parent application must be shipped with root metadata, and must
        # ensure it is placed in the metadata_dir (without version in filename)
        shutil.copy(
            src=TEST_REPO_DIR / 'metadata' / ('1.' + ROOT_FILENAME),
            dst=self.metadata_dir / ROOT_FILENAME,
        )
        # kwargs for client initializer
        self.client_kwargs = dict(
            app_name='example_app',
            app_install_dir=self.app_install_dir,
            current_version='1.0',
            metadata_dir=self.metadata_dir,
            metadata_base_url='http://localhost:8000/metadata/',
            target_dir=self.target_dir,
            target_base_url='http://localhost:8000/targets/',
            session_auth={'http://localhost:8000': ('username', 'password')},
        )

    def mock_download_metadata(
        self, rolename: str, length: int, version: Optional[int] = None
    ) -> bytes:
        if rolename == 'root':
            # indicate current root is newest version
            raise tuf.api.exceptions.DownloadHTTPError(status_code=404, message='')
        # read from the test repo dir, instead of actually downloading
        filename = f'{rolename}.json'
        if version:
            filename = f'{version}.{filename}'
        file_path = TEST_REPO_DIR / 'metadata' / filename
        return file_path.read_bytes()

    def get_refreshed_client(self):
        # refresh to load targets metadata (mock to prevent actual download)
        client = Client(**self.client_kwargs)
        with patch.object(client, '_download_metadata', self.mock_download_metadata):
            client.refresh()
        # ensure current archive exists (dummy)
        shutil.copy(
            src=TEST_REPO_DIR / 'targets' / client.current_archive.path.name,
            dst=client.current_archive_local_path,
        )
        return client

    def test_init_no_metadata(self):
        # cannot initialize without root metadata file
        (self.metadata_dir / ROOT_FILENAME).unlink()
        with self.assertRaises(FileNotFoundError):
            Client(**self.client_kwargs)

    def test_init(self):
        client = Client(**self.client_kwargs)
        # Client is a subclass of tuf.ngclient.Updater, so it automatically
        # loads and verifies the local root metadata file
        self.assertTrue(client._trusted_set.root)
        # other metadata is not available yet
        for role_name in ['targets', 'snapshot', 'timestamp']:
            # see python-tuf #2250
            self.assertNotIn(role_name, client._trusted_set)

    def test_trusted_target_metas(self):
        client = self.get_refreshed_client()
        self.assertTrue(client.trusted_target_metas)
        # The archives have custom metadata from the user, as defined in the
        # repo_workflow_example.py script. The patches have custom metadata for .tar
        # integrity check (internal).
        for meta in client.trusted_target_metas:
            with self.subTest(msg=meta):
                if str(meta.version) != '1.0':
                    if meta.is_archive:
                        self.assertTrue(meta.custom)
                        example_key = 'changes'  # see repo workflow example
                        self.assertIn(example_key, meta.custom)
                    else:
                        # patches must have tar hash information
                        self.assertIn('tar_hash', meta.custom_internal)

    def test_get_targetinfo(self):
        client = self.get_refreshed_client()
        target_path_str = 'example_app-1.0.tar.gz'
        target_meta = TargetMeta(target_path=target_path_str)
        for target_path in [target_path_str, target_meta]:
            target_info = client.get_targetinfo(target_path=target_path)
            with self.subTest(msg=target_path):
                self.assertIsInstance(target_info, TargetFile)

    def test_updates_available(self):
        client = Client(**self.client_kwargs)
        # test check_for_updates not called
        with self.assertLogs(logger='tufup.client', level=logging.WARNING) as cm:
            self.assertFalse(client.updates_available)
        self.assertIn('check_for_updates', cm.output[0])
        # test check_for_updates called and update available
        client.new_targets = {'dummy': None}
        self.assertTrue(client.updates_available)

    def test_download_and_apply_update(self):
        # just for completeness...
        mock_download = Mock(return_value=True)
        mock_apply = Mock(return_value=True)
        mock_install = Mock()
        with patch.multiple(
            Client, _download_updates=mock_download, _apply_updates=mock_apply
        ):
            client = self.get_refreshed_client()
            client.new_targets = {'dummy': None}
            client.download_and_apply_update(install=mock_install)
        self.assertTrue(mock_download.called)
        self.assertTrue(mock_apply.called)
        self.assertIn(mock_install, mock_apply.call_args.kwargs.values())

    def test_check_for_updates(self):
        client = self.get_refreshed_client()
        with patch.object(client, 'refresh', Mock()):
            # expectations (based on targets in tests/data/repository, but ignoring
            # required releases):
            # - pre=None, '', or 'invalid':
            #     only full releases are included, so this finds 2.0 (patch)
            # - pre='a': finds all, but total patch size exceeds archive size
            # - pre='b': there is no 'b' release, so this finds same as 'rc'
            # - pre='rc': finds 2.0 and 3.0rc0, total patch size smaller than archive
            for pre, expected in [
                (None, 1),
                ('', 1),
                ('a', 1),
                ('b', 2),
                ('rc', 2),
                ('invalid', 1),
            ]:
                with self.subTest(msg=pre):
                    # verify that we always find the required release, unless we
                    # explicitly set ignore_required=True
                    required_version = '2.0'
                    target_meta = client.check_for_updates(pre=pre)
                    self.assertEqual(required_version, str(target_meta.version))
                    # for the actual test we want to treat all versions as not-required
                    target_meta = client.check_for_updates(
                        pre=pre, ignore_required=True
                    )
                    self.assertTrue(expected and target_meta)
                    self.assertEqual(expected, len(client.new_targets))
                    self.assertTrue(
                        all(
                            getattr(item, 'is_archive' if pre == 'a' else 'is_patch')
                            for item in client.new_targets.keys()
                        )
                    )
                    # verify that we can access custom metadata where needed
                    if target_meta.is_patch:
                        self.assertTrue(target_meta.custom)

    def test_check_for_updates_already_up_to_date(self):
        self.client_kwargs['current_version'] = '4.0a0'
        client = self.get_refreshed_client()
        with patch.object(client, 'refresh', Mock()):
            self.assertFalse(client.check_for_updates(pre='a'))

    def test_check_for_updates_current_archive_missing(self):
        client = self.get_refreshed_client()
        # remove current archive dummy
        client.current_archive_local_path.unlink()
        with patch.object(client, 'refresh', Mock()):
            for pre in [None, 'a', 'b', 'rc']:
                self.assertTrue(client.check_for_updates(pre=pre))
                target_meta = next(iter(client.new_targets.keys()))
                self.assertTrue(target_meta.is_archive)

    def test_check_for_updates_failed_patch(self):
        client = self.get_refreshed_client()
        # first verify that we would normally get a patch update
        with patch.object(client, 'refresh', Mock()):
            client.check_for_updates()
            target_meta = next(iter(client.new_targets.keys()))
            self.assertTrue(target_meta.is_patch)
        # copy the patch into the client cache to simulate existing failed patch
        client_cache_dir = pathlib.Path(client.target_dir)
        shutil.copy(
            src=TEST_REPO_DIR / 'targets' / target_meta.filename,
            dst=client_cache_dir / (target_meta.filename + SUFFIX_FAILED),
        )
        # test: should fall back on full archive update
        with patch.object(client, 'refresh', Mock()):
            with self.assertLogs(level='DEBUG') as logs:
                client.check_for_updates()
                target_meta = next(iter(client.new_targets.keys()))
            self.assertTrue(target_meta.is_archive)
            self.assertIn('aborting', ''.join(logs.output))

    def test_check_for_updates_after_patch(self):
        """
        Our patching process ensures that the reconstructed uncompressed .tar archive
        is identical to the uncompressed original. However, the gzip compressed
        archives are likely to be different (gzip output is not guaranteed to be
        identical, between implementations, even for identical input). This means the
        hash of a *reconstructed* .tar.gz file most likely does not match the hash of
        the original .tar.gz file (as found in the tuf metadata). This test verifies
        that the tuf target hash mismatch does not impede the patch update process.
        """
        client = self.get_refreshed_client()
        # copy current compressed archive for reference
        original_gz = client.current_archive_local_path.with_name('original.tar.gz')
        shutil.copy(src=client.current_archive_local_path, dst=original_gz)
        # decompress and recompress archive, as if it were reconstructed from patch
        with gzip.open(original_gz, 'rb') as src:
            with gzip.open(client.current_archive_local_path, 'wb') as dst:
                dst.write(src.read())
        # verify that the two .tar.gz files are indeed different (even though the
        # uncompressed .tar content is identical)
        self.assertFalse(
            filecmp.cmp(original_gz, client.current_archive_local_path, shallow=False)
        )
        # test: even though the .tar.gz hash does not match the one in the tuf
        # metadata, a patch update should be scheduled
        with patch.object(client, 'refresh', Mock()):
            archive_meta = client.check_for_updates()
            # check_for_updates always returns *archive* meta (or None), even if the
            # actual update is a *patch*
            self.assertTrue(archive_meta)
            # verify that the update is a *patch*
            target_meta = next(iter(client.new_targets.keys()))
            self.assertTrue(target_meta.is_patch)

    def test__download_updates(self):
        client = Client(**self.client_kwargs)
        client.new_targets = {Mock(): Mock()}
        for cached_path, downloaded_path in [('cached', None), (None, 'downloaded')]:
            with patch.multiple(
                client,
                find_cached_target=Mock(return_value=cached_path),
                download_target=Mock(return_value=downloaded_path),
            ):
                self.assertTrue(client._download_updates(progress_hook=None))
                local_path = next(iter(client.downloaded_target_files.values()))
                if cached_path:
                    self.assertEqual(cached_path, str(local_path))
                else:
                    self.assertEqual(downloaded_path, str(local_path))

    def test__apply_updates(self):
        client = self.get_refreshed_client()
        # copy files from test data to temporary client cache
        client.downloaded_target_files = dict()
        for target_meta in client.trusted_target_metas:
            if target_meta.is_patch and str(target_meta.version) in ['2.0', '3.0rc0']:
                src_path = TEST_REPO_DIR / 'targets' / target_meta.filename
                dst_path = pathlib.Path(client.target_dir, target_meta.filename)
                shutil.copy(src=src_path, dst=dst_path)
                client.downloaded_target_files[target_meta] = dst_path
        # specify new archive (normally done in _check_updates)
        archives = [
            tp
            for tp in client.trusted_target_metas
            if tp.is_archive and str(tp.version) == '3.0rc0'
        ]
        client.new_archive_info = client.get_targetinfo(archives[-1])
        client.new_archive_local_path = pathlib.Path(
            client.target_dir, client.new_archive_info.path
        )
        # tests
        with self.subTest(msg='with confirmation'):
            mock_install = Mock()
            with patch('builtins.input', Mock(return_value='y')):
                client._apply_updates(install=mock_install, skip_confirmation=False)
            self.assertTrue(any(client.extract_dir.iterdir()))
            self.assertTrue(mock_install.called)
        with self.subTest(msg='skip confirmation'):
            mock_install = Mock()
            client._apply_updates(install=mock_install, skip_confirmation=True)
            mock_install.assert_called()
        with self.subTest(msg='patch failure due to mismatch'):
            mock_install = Mock()
            with patch.object(
                tufup.client.Patcher,
                '_verify_tar_size_and_hash',
                Mock(side_effect=Exception()),
            ):
                client._apply_updates(install=mock_install, skip_confirmation=True)
            mock_install.assert_not_called()
            target_paths = pathlib.Path(client.target_dir).iterdir()
            self.assertTrue(any(path.suffix == SUFFIX_FAILED for path in target_paths))

    def test_version_comparison(self):
        # verify assumed version hierarchy
        v = packaging.version.Version
        self.assertTrue(
            v('0.0.0') < v('0.1a') < v('0.1b') < v('0.1rc') < v('0.1rc1') < v('0.1')
        )
        self.assertEqual(v('0.1rc'), v('0.1rc0'))


class AuthRequestsFetcherTests(unittest.TestCase):
    def setUp(self) -> None:
        self.session_auth = {
            'https://example.net': None,
            'https://example.com': ('username', 'password'),
            'https://example.org': HTTPBasicAuth(username='x', password='y'),
            'http://localhost:8000': ('username', 'password'),
        }

    def test_init(self):
        # drop-in replacement for default RequestsFetcher, without args
        self.assertTrue(AuthRequestsFetcher())
        # if authentication is required, specify arg
        fetcher = AuthRequestsFetcher(session_auth=self.session_auth)
        self.assertEqual(self.session_auth, fetcher.session_auth)

    def test__get_session(self):
        fetcher = AuthRequestsFetcher(session_auth=self.session_auth)
        for scheme_and_server, auth in self.session_auth.items():
            url = scheme_and_server + '/some/path?query=something'
            with self.subTest(msg=url):
                session = fetcher._get_session(url=url)
                self.assertEqual(auth, session.auth)

    @unittest.skipIf(condition=ON_GITHUB, reason='external dependency')
    def test_fetch_basic_auth(self):
        # kind of an integration test, as it connects to an external server...
        scheme_and_server = 'https://httpbin.org'
        user = 'me'
        passwd = 'mypassword'
        url = f'{scheme_and_server}/basic-auth/{user}/{passwd}'
        session_auth = {
            scheme_and_server: HTTPBasicAuth(username=user, password=passwd)
        }
        fetcher = AuthRequestsFetcher(session_auth=session_auth)
        fetcher.socket_timeout = 30  # in case httpbin.org is slow to respond
        # we don't have direct access to the response, so we'll just check
        # that RequestsFetcher.fetch() doesn't raise an error, such as a
        # status "401 Unauthorized" or "403 Forbidden"
        try:
            fetcher.fetch(url=url)
        except tuf.api.exceptions.DownloadHTTPError as e:
            self.fail(msg=f'fetch() raised unexpected HTTPError: {e}')

    def test_attach_progress_hook(self):
        mock_hook = Mock()
        bytes_expected = 10
        fetcher = AuthRequestsFetcher()
        fetcher.attach_progress_hook(hook=mock_hook, bytes_expected=bytes_expected)
        bytes_new = 1
        bytes_downloaded = 0
        while bytes_downloaded < bytes_expected:
            bytes_downloaded += bytes_new
            fetcher._progress(bytes_new=bytes_new)
            mock_hook.assert_called_with(
                bytes_downloaded=bytes_downloaded, bytes_expected=bytes_expected
            )

    def test__chunks_without_progress_hook(self):
        chunk_size = 10
        chunk_count = 10
        chunks = [b'x' * chunk_size] * chunk_count

        def mock_iter_content(*args):
            yield from chunks

        mock_response = Mock(iter_content=mock_iter_content, close=Mock())
        fetcher = AuthRequestsFetcher()
        fetcher.chunk_size = chunk_size
        # _chunks should work even if attach_progress_hook was not called
        try:
            for __ in fetcher._chunks(response=mock_response):
                pass
        except Exception as e:
            self.fail(msg=f'_chunks raised an unexpected exception: {e}')

    def test__chunks_with_progress_hook(self):
        chunk_size = 10
        chunk_count = 10
        chunks = [b'x' * chunk_size] * chunk_count

        def mock_iter_content(*args):
            yield from chunks

        mock_response = Mock(iter_content=mock_iter_content, close=Mock())
        fetcher = AuthRequestsFetcher()
        fetcher.chunk_size = chunk_size
        # test custom progress hook
        mock_hook = Mock()
        bytes_expected = chunk_size * chunk_count
        fetcher.attach_progress_hook(hook=mock_hook, bytes_expected=bytes_expected)
        for __ in fetcher._chunks(response=mock_response):
            pass
        self.assertEqual(chunk_count, mock_hook.call_count)



import pathlib
from getpass import getuser
import os
import subprocess
import sys
import tempfile
import textwrap
from time import sleep
import unittest
from unittest.mock import patch

from tests import BASE_DIR, TempDirTestCase
import tufup.utils.platform_specific as ps
from tufup.utils.platform_specific import (
    ON_MAC,
    ON_WINDOWS,
    PLATFORM_SUPPORTED,
    run_bat_as_admin,
    WIN_BATCH_PREFIX,
    WIN_BATCH_SUFFIX,
)

_reason_platform_not_supported = (
    'install_update() is only actively supported on windows and mac'
)

DUMMY_APP_CONTENT = f"""
import subprocess
import sys
sys.path.append('{(BASE_DIR.parent / 'src').as_posix()}')
from tufup.utils.platform_specific import install_update
install_update(src_dir=sys.argv[1], dst_dir=sys.argv[2], {{extra_kwargs_str}})
"""

ON_GITHUB = os.getenv('GITHUB_ACTIONS')
TEST_RUNAS = os.getenv('TEST_RUNAS')


class UtilsTests(TempDirTestCase):
    def setUp(self) -> None:
        super().setUp()
        # create src dir with dummy app file, and dst dir with stale subdir
        # and a file that must be excluded from purge
        test_dir = self.temp_dir_path / 'tufup_tests'
        self.src_dir = test_dir / 'src'
        self.src_subdir = self.src_dir / 'new'
        self.src_subdir.mkdir(parents=True)
        self.dst_dir = test_dir / 'dst'
        self.dst_subdir = self.dst_dir / 'stale'
        self.dst_subdir.mkdir(parents=True)
        (self.dst_subdir / 'stale.file').touch()
        self.keep_file_path = self.dst_dir / 'keep.file'
        self.keep_file_path.touch()
        self.keep_file_str = str(self.keep_file_path).replace('\\', '\\\\')
        self.src_file_name = 'dummy_app.py'
        self.src_file_path = self.src_dir / self.src_file_name

    def run_dummy_app(self, extra_kwargs_strings):
        # write dummy app content to file
        dummy_app_content = DUMMY_APP_CONTENT.format(
            extra_kwargs_str=', '.join(extra_kwargs_strings),
        )
        print(dummy_app_content)
        self.src_file_path.write_text(dummy_app_content)
        # run the dummy app in a separate process, which, in turn, will run
        # another process that moves the file
        completed_process = subprocess.run(
            [sys.executable, self.src_file_path, self.src_dir, self.dst_dir]
        )
        print(sys.executable)
        completed_process.check_returncode()
        if ON_WINDOWS:
            # allow some time for the batch file to complete (the batch file
            # waits a few seconds, so we have to wait longer)
            sleep(3)

    @unittest.skipIf(
        condition=ON_GITHUB or not TEST_RUNAS or not ON_WINDOWS,
        reason='windows only, requires user interaction',
    )
    def test_run_bat_as_admin(self):
        output_path = self.temp_dir_path / 'output.txt'
        bat_path = self.temp_dir_path / 'tell_me_who_i_am.bat'
        bat_path.write_text(f'whoami > "{output_path}"\ntimeout /t -1')
        # NOTE: this will open a UAC prompt (User Access Control)
        self.assertTrue(run_bat_as_admin(file_path=bat_path))
        # doesn't block, so we'll pause for a while
        sleep(1)
        self.assertTrue(output_path.exists())
        output = output_path.read_text()
        current_user = getuser()
        print(f'bat file runs as: {output}')
        print(f'current user: {current_user}')
        self.assertTrue(len(output))
        self.assertNotIn(current_user, output)

    @unittest.skipIf(condition=not ON_MAC, reason='macOS only')
    def test_install_update_macos_symlinks(self):
        with patch.object(ps, '_install_update_mac') as mock_install_update_mac:
            ps.install_update(src_dir='', dst_dir='')
            self.assertNotIn('symlinks', mock_install_update_mac.call_args.kwargs)
            ps.install_update(src_dir='', dst_dir='', symlinks=True)
            self.assertTrue(mock_install_update_mac.call_args.kwargs['symlinks'])

    @unittest.skipIf(
        condition=not PLATFORM_SUPPORTED, reason=_reason_platform_not_supported
    )
    def test_install_update_no_purge(self):
        extra_kwargs_strings = []
        if ON_WINDOWS:
            extra_kwargs_strings.extend(
                # unknown_kwarg reproduces issue #126
                ['as_admin=False', 'log_file_name=None', 'unknown_kwarg=True']
            )
        # run the dummy app in a separate process
        self.run_dummy_app(extra_kwargs_strings=extra_kwargs_strings)
        # ensure file has been moved from src to dst
        self.assertTrue(any(self.dst_dir.iterdir()))
        self.assertTrue((self.dst_dir / self.src_file_name).exists())
        # new empty subdir has been moved as well
        self.assertTrue((self.dst_dir / self.src_subdir.name).exists())
        # original src file no longer exists
        self.assertFalse(self.src_file_path.exists())
        # stale dst content must still be present
        self.assertTrue(self.dst_subdir.exists())
        # file to keep must still be present
        self.assertTrue(self.keep_file_path.exists())
        # the batch file itself should have been removed
        self.assertFalse(batch_file_exists())

    @unittest.skipIf(
        condition=not PLATFORM_SUPPORTED, reason=_reason_platform_not_supported
    )
    def test_install_update_purge(self):
        extra_kwargs_strings = [
            'purge_dst_dir=True',
            f'exclude_from_purge=["{self.keep_file_str}"]',
        ]
        if ON_WINDOWS:
            extra_kwargs_strings.extend(['as_admin=False', 'log_file_name=None'])
        # run the dummy app in a separate process
        self.run_dummy_app(extra_kwargs_strings=extra_kwargs_strings)
        # ensure file has been moved from src to dst
        self.assertTrue(any(self.dst_dir.iterdir()))
        self.assertTrue((self.dst_dir / self.src_file_name).exists())
        # new empty subdir has been moved as well
        self.assertTrue((self.dst_dir / self.src_subdir.name).exists())
        # original src file no longer exists
        self.assertFalse(self.src_file_path.exists())
        # stale dst content has been removed (robocopy /purge)
        self.assertFalse(self.dst_subdir.exists())
        # file to keep must still be present
        self.assertTrue(self.keep_file_path.exists())

    @unittest.skipIf(condition=not ON_WINDOWS, reason='robocopy is windows only')
    def test_install_update_robocopy_options_override(self):
        extra_kwargs_strings = [
            'as_admin=False',
            'log_file_name=None',
            'robocopy_options_override=[]',
        ]
        # run the dummy app in a separate process
        self.run_dummy_app(extra_kwargs_strings=extra_kwargs_strings)
        # ensure file has been copied from src to dst
        self.assertTrue(any(self.dst_dir.iterdir()))
        self.assertTrue((self.dst_dir / self.src_file_name).exists())
        # new subdir has not been copied
        self.assertFalse((self.dst_dir / self.src_subdir.name).exists())
        # original src file still exists
        self.assertTrue(self.src_file_path.exists())
        # stale dst content must still be present
        self.assertTrue(self.dst_subdir.exists())
        # file to keep must still be present
        self.assertTrue(self.keep_file_path.exists())

    @unittest.skipIf(
        condition=not ON_WINDOWS, reason='install.log file is windows only'
    )
    def test_install_update_log_file(self):
        log_file_name = 'install.log'
        extra_kwargs_strings = [
            'as_admin=False',
            f'log_file_name="{log_file_name}"',
            'robocopy_options_override=[]',
        ]
        # run the dummy app in a separate process
        self.run_dummy_app(extra_kwargs_strings=extra_kwargs_strings)
        # a log file should exist
        log_file_path = self.dst_dir / log_file_name
        self.assertTrue(log_file_path.exists())
        log_file_content = log_file_path.read_text()
        self.assertTrue(log_file_content)

    @unittest.skipIf(
        condition=not ON_WINDOWS, reason='process_creation_flags is for windows only'
    )
    def test_install_update_process_creation_flags(self):
        # the log file is only used to verify that the batch file has run successfully
        log_file_name = 'install.log'
        extra_kwargs_strings = [
            'process_creation_flags=subprocess.CREATE_NO_WINDOW',
            f'log_file_name="{log_file_name}"',
        ]
        # run the dummy app in a separate process
        self.run_dummy_app(extra_kwargs_strings=extra_kwargs_strings)
        # a log file should exist
        log_file_path = self.dst_dir / log_file_name
        self.assertTrue(log_file_path.read_text())

    @unittest.skipIf(
        condition=not ON_WINDOWS, reason='windows batch files are windows only'
    )
    def test_install_update_custom_batch_template(self):
        custom_content = 'some custom text'
        custom_file_path = self.temp_dir_path / 'test.txt'
        # a custom batch template that ignores most of the default template
        # variables and adds some custom variables
        custom_batch_template = textwrap.dedent(
            """
            echo {custom_content}> "{custom_file_path}"
            {delete_self}
            """
        )
        extra_kwargs_strings = [
            f'batch_template="""{custom_batch_template}"""',
            f'batch_template_extra_kwargs=dict(custom_content="{custom_content}", custom_file_path=r"{custom_file_path}")',
        ]
        # run the dummy app in a separate process
        self.run_dummy_app(extra_kwargs_strings=extra_kwargs_strings)
        # batch file should have created a file with specified content
        self.assertTrue(custom_file_path.exists())
        custom_file_content = custom_file_path.read_text()
        self.assertIn(custom_content, custom_file_content)
        # the batch file itself should have been removed
        self.assertFalse(batch_file_exists())


def batch_file_exists():
    """
    Checks if any tufup batch files remain in the system temporary directory.

    BEWARE: If a batch file does not remove itself successfully even once,
    this function will keep returning True, until the batch file is removed
    manually from the system temp dir (see print statement below). Windows
    does not clear the temp dirs automatically.
    """
    system_temp_dir = pathlib.Path(tempfile.gettempdir())
    print(f'system temp dir: {system_temp_dir}')
    return any(
        path.name.startswith(WIN_BATCH_PREFIX) and path.name.endswith(WIN_BATCH_SUFFIX)
        for path in system_temp_dir.iterdir()
        if path.is_file()
    )



from copy import deepcopy
import logging
import pathlib
import shutil
import sys
import tempfile
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Union
from urllib import parse

import requests
from requests.auth import AuthBase
from tuf.api.exceptions import DownloadError, UnsignedMetadataError
import tuf.ngclient

from tufup.common import BinaryDiff, KEY_REQUIRED, Patcher, TargetMeta
from tufup.utils.platform_specific import install_update

logger = logging.getLogger(__name__)

DEFAULT_EXTRACT_DIR = pathlib.Path(tempfile.gettempdir()) / 'tufup'
SUFFIX_FAILED = '.failed'
# do full update if patch-size/full-size > MAX_SIZE_RATIO
MAX_SIZE_RATIO = 0.8


class Client(tuf.ngclient.Updater):
    def __init__(
        self,
        app_name: str,
        app_install_dir: pathlib.Path,
        current_version: str,
        metadata_dir: pathlib.Path,
        metadata_base_url: str,
        target_dir: pathlib.Path,
        target_base_url: str,
        extract_dir: Optional[pathlib.Path] = None,
        refresh_required: bool = False,
        session_auth: Optional[Dict[str, Union[Tuple[str, str], AuthBase]]] = None,
        binary_diff: Optional[type[BinaryDiff]] = None,
        **kwargs,
    ):
        """
        The `tufup.client.Client` is a subclass of `tuf.ngclient.Updater`.

        for `session_auth` formats, see docs for `tufup.client.AuthRequestsFetcher`
        """
        # tuf.ngclient.Updater.__init__ loads local root metadata automatically
        super().__init__(
            metadata_dir=str(metadata_dir),
            metadata_base_url=metadata_base_url,
            target_dir=str(target_dir),
            target_base_url=target_base_url,
            fetcher=AuthRequestsFetcher(session_auth=session_auth),
            **kwargs,
        )
        self.app_install_dir = app_install_dir
        self.extract_dir = extract_dir
        self.refresh_required = refresh_required
        self.current_archive = TargetMeta(name=app_name, version=current_version)
        self.current_archive_local_path = target_dir / self.current_archive.path
        self.new_archive_local_path: Optional[pathlib.Path] = None
        self.new_archive_info: Optional[tuf.ngclient.TargetFile] = None
        self.new_targets: Optional[dict] = None
        self.downloaded_target_files = {}
        self.binary_diff = binary_diff

    @property
    def trusted_target_metas(self) -> list:
        """
        Return a list of available trusted targets, as TargetMeta objects.

        This is convenient because TargetMeta objects can be sorted by version.
        """
        # todo: _trusted_set is private, but ideally we would use a public
        #  interface (if only tuf.ngclient exposed one...)
        _trusted_target_metas = []
        if self._trusted_set.targets:
            _trusted_target_metas = [
                TargetMeta(target_path=key, custom=target.custom)
                for key, target in self._trusted_set.targets.targets.items()
            ]
            logger.debug(f'{len(_trusted_target_metas)} TargetMeta objects created')
        else:
            logger.warning('targets metadata not found')
        return _trusted_target_metas

    def get_targetinfo(
        self, target_path: Union[str, TargetMeta]
    ) -> Optional[tuf.ngclient.TargetFile]:
        """Extend Updater.get_targetinfo to handle TargetMeta input args."""
        if isinstance(target_path, TargetMeta):
            target_path = target_path.target_path_str
        return super().get_targetinfo(target_path=target_path)

    @property
    def updates_available(self):
        if self.new_targets is None:
            logger.warning('Please call check_for_updates first.')
            return False
        else:
            return len(self.new_targets) > 0

    def download_and_apply_update(
        self,
        skip_confirmation: bool = False,
        install: Optional[Callable] = None,
        progress_hook: Optional[Callable] = None,
        **kwargs,
    ):
        """
        Download and apply available updates.

        Note that `check_for_updates` must be called first.

        This downloads the files found by `check_for_updates`, applies any
        patches, and extracts the resulting archive to the `extract_dir`. At
        that point, the update is ready to be installed (i.e. moved into
        place). This is done by calling `install` with the specified `**kwargs`.

        The default `install` callable moves the content of `extract_dir` to
        `app_install_dir`, and exits the application (not necessarily in that
        order).

        The **kwargs are passed on to the 'install' callable

        The default `install` callable accepts two additional arguments:

            `purge_dst_dir` (default False): if True, *ALL* content will be
            deleted from the `app_install_dir`

            `exclude_from_purge` (default None): list of paths to exclude
            from purge

            **DANGER**: Only set `purge_dst_dir=True` if your app is
            installed in its own separate directory, otherwise this will
            cause unrelated files and folders to be deleted.
        """
        if install is None:
            install = install_update
        if self.updates_available and self._download_updates(
            progress_hook=progress_hook
        ):
            self._apply_updates(
                install=install, skip_confirmation=skip_confirmation, **kwargs
            )

    def check_for_updates(
        self,
        pre: Optional[str] = None,
        patch: bool = True,
        ignore_required: bool = False,
    ) -> Optional[TargetMeta]:
        """
        Check if any updates are available, based on current app version.

        Returns latest archive meta, if a new archive is found.

        Final releases are always included. Pre-releases are excluded by
        default. If `pre` is specified, pre-releases are included, down to
        the specified level. Pre-release identifiers follow the PEP440
        specification, i.e. 'a', 'b', or 'rc', for alpha, beta, and release
        candidate, respectively.

        If `patch` is `False`, a full update is enforced.

        If a new release is marked as "required" (in its custom metadata) this
        release will take precedence over any non-required releases, *even* if the
        latter are newer. This may be useful e.g. in case of a configuration change.
        These "required" releases should be rare, and should preferably be avoided.
        However, in the exceedingly rare event that there *are* "required" updates,
        yet the user wants to treat them as non-required, they can specify
        `ignore_required=True`.
        """
        # invalid pre-release specifiers are ignored, with a warning
        pre_map = dict(a='abrc', b='brc', rc='rc')
        prereleases = pre_map.get(pre, '')
        if pre and not prereleases:
            logger.warning(f'ignoring invalid pre-release specifier: "{pre}"')
        # refresh top-level metadata (root -> timestamp -> snapshot -> targets)
        try:
            self.refresh()
        except (DownloadError, UnsignedMetadataError) as e:
            logger.warning(f'Cannot refresh metadata: {e}')
            if self.refresh_required:
                logger.warning('Exiting: refresh is required')
                sys.exit()
            return None
        # check for new target files (archives and patches)
        logger.debug(f'current archive: {self.current_archive.filename}')
        all_new_targets = dict(
            (target_meta, self.get_targetinfo(target_meta))
            for target_meta in self.trusted_target_metas
            if target_meta.name == self.current_archive.name
            and target_meta.version > self.current_archive.version
        )
        logger.debug(f'{len(all_new_targets)} new *targets* found')
        # determine latest archive, filtered by the specified pre-release level
        new_archives = dict(
            item
            for item in all_new_targets.items()
            if item[0].is_archive
            and (not item[0].version.pre or item[0].version.pre[0] in prereleases)
        )
        new_archive_meta = None
        if new_archives:
            logger.debug(f'{len(new_archives)} new *archives* found')
            # the "latest" archive is typically just the last one in the sorted list
            # of new archives, except when there are new "required" archives,
            # in which case we must update to the first "required" archive encountered
            for archive_meta, archive_info in sorted(new_archives.items()):
                if not ignore_required and archive_meta.custom_internal:
                    if archive_meta.custom_internal.get(KEY_REQUIRED):
                        logger.debug(f'required update found: {archive_meta.version}')
                        break
            new_archive_meta = archive_meta  # noqa
            self.new_archive_info = archive_info  # noqa
            self.new_archive_local_path = pathlib.Path(
                self.target_dir, new_archive_meta.path.name
            )
            # patches must include all pre-releases and final releases up to,
            # and including, the latest archive as determined above
            new_patches = dict(
                item
                for item in all_new_targets.items()
                if item[0].is_patch and item[0].version <= new_archive_meta.version
            )
            # determine size of patch update and archive update
            total_patch_size = sum(
                target_file.length for target_file in new_patches.values()
            )
            # abort patch update if any of the new patches have failed on a previous run
            abort_patch = False
            for patch_info in new_patches.values():
                patch_info_mod = deepcopy(patch_info)  # modify a copy, just to be sure
                patch_info_mod.path += SUFFIX_FAILED
                if self.find_cached_target(targetinfo=patch_info_mod):
                    logger.debug(f'aborting patch due to {patch_info_mod.path}')
                    abort_patch = True
            # use file size to decide if we want to do a patch update or a
            # full update (if there are no patches, or if the current archive
            # is not available, we must do a full update)
            self.new_targets = new_patches
            no_patches = total_patch_size == 0
            patch_too_big = (
                total_patch_size / self.new_archive_info.length > MAX_SIZE_RATIO
            )
            no_archive = not self.current_archive_local_path.exists()
            if not patch or no_patches or patch_too_big or no_archive or abort_patch:
                # fall back on full update
                self.new_targets = {new_archive_meta: self.new_archive_info}
                logger.debug('full update available')
            else:
                # continue with patch update
                logger.debug('patch update(s) available')
        else:
            self.new_targets = {}
            logger.debug('no new archives found')
        return new_archive_meta

    def _download_updates(self, progress_hook: Optional[Callable]) -> bool:
        # download the new targets selected in check_for_updates
        for target_meta, target_file in self.new_targets.items():
            # check if the target file has already been downloaded
            local_path_str = self.find_cached_target(targetinfo=target_file)
            if not local_path_str:
                # attach progress hook
                if progress_hook:
                    self._fetcher.attach_progress_hook(
                        hook=progress_hook, bytes_expected=target_file.length
                    )
                # download the target file
                local_path_str = self.download_target(targetinfo=target_file)
            self.downloaded_target_files[target_meta] = pathlib.Path(local_path_str)
        return len(self.downloaded_target_files) == len(self.new_targets)

    def _apply_updates(
        self,
        install: Callable,
        skip_confirmation: bool,
        **kwargs,
    ):
        """
        kwargs are passed on to the 'install' callable

        Note this has a side-effect: if self.extract_dir is not specified,
        an extract_dir is created in a platform-specific temporary location.
        """
        # either patch the current archive (if we have patches) or use new full archive
        try:
            if next(iter(self.downloaded_target_files.keys())).is_archive:
                # full archive is available
                if len(self.downloaded_target_files) != 1:
                    raise ValueError('there should be only one downloaded *archive*')
                if not self.new_archive_local_path.exists():
                    raise FileNotFoundError('the new archive file does not exist')
            else:
                # reconstruct full archive from patch(es)
                if not all(
                    target.is_patch for target in self.downloaded_target_files.keys()
                ):
                    raise ValueError('all downloaded targets must be patches')
                Patcher.patch_and_verify(
                    src_path=self.current_archive_local_path,
                    dst_path=self.new_archive_local_path,
                    patch_targets=self.downloaded_target_files,
                    binary_diff=self.binary_diff,
                )
        except Exception as e:
            # rename all failed targets in order to skip them (patches) or retry
            # them (archive) on the next run
            for target, file_path in self.downloaded_target_files.items():
                renamed_path = file_path.replace(
                    file_path.with_suffix(file_path.suffix + SUFFIX_FAILED)
                )
                logger.debug(f'update failed: target renamed to {renamed_path}')
            logger.error(f'update aborted: {e}')
            return
        # extract archive to temporary directory
        if self.extract_dir is None:
            self.extract_dir = DEFAULT_EXTRACT_DIR
            self.extract_dir.mkdir(exist_ok=True)
            logger.debug(f'default extract dir created: {self.extract_dir}')
        # extract
        shutil.unpack_archive(
            filename=self.new_archive_local_path, extract_dir=self.extract_dir
        )
        logger.debug(f'files extracted to {self.extract_dir}')
        # install
        confirmation_message = f'Install update in {self.app_install_dir}? [y]/n'
        if skip_confirmation or input(confirmation_message) in ['y', '']:
            # start a script that moves the extracted files to the app install
            # directory (overwrites existing files), then exit current process
            install(
                src_dir=self.extract_dir,
                dst_dir=self.app_install_dir,
                **kwargs,
            )
        else:
            logger.warning('Installation aborted.')
        # todo: clean up deprecated local archive


class AuthRequestsFetcher(tuf.ngclient.RequestsFetcher):
    # RequestsFetcher is public as of python-tuf v2.1.0 (see python-tuf #2277)
    def __init__(
        self,
        session_auth: Optional[Dict[str, Union[Tuple[str, str], AuthBase]]] = None,
    ) -> None:
        """
        This extends the default tuf RequestsFetcher, so we can specify
        authentication tuples (or custom authentication objects) for each
        session.

        session_auth (optional):

            dict of the form {<scheme and server>: (<username>, <password>), ...}
            or {<scheme and server>: <requests.auth.AuthBase>, ...}
            or some combination of those

        where <scheme and server> can be e.g. https://example.com
        or http://localhost:8000.

        Note: <scheme and server> must *not* have a trailing slash

        Naming follows [RFC 2396][1], which defines a generic uri as:

            <scheme>://<authority><path>?<query>

        where <authority> can be <server>.

        Also see session authentication example in requests docs: [1][2][3]

        [1]: https://datatracker.ietf.org/doc/html/rfc2396#section-3
        [2]: https://docs.python-requests.org/en/master/user/advanced/#session-objects
        [3]: https://docs.python-requests.org/en/latest/user/advanced/#custom-authentication
        [4]: https://docs.python-requests.org/en/master/api/#sessionapi
        """
        super().__init__()
        self.session_auth = session_auth or {}
        # default progress hook does nothing
        self._progress = lambda bytes_new: None

    def attach_progress_hook(self, hook: Callable, bytes_expected: int):
        """
        Allow clients to attach a progress hook which gets called after every
        downloaded chunk.

        The hook must accept two kwargs: bytes_downloaded and bytes_expected
        """

        def progress(
            bytes_new: int,
            _cache: List[int] = [],  # noqa: mutable default intentional
        ):
            # mutable default is used to keep track of downloaded chunk sizes
            _cache.append(bytes_new)
            return hook(bytes_downloaded=sum(_cache), bytes_expected=bytes_expected)

        self._progress = progress

    def _get_session(self, url: str) -> requests.Session:
        # set the Session.auth attribute for the specified server, if available
        session = super()._get_session(url=url)
        # determine session_auth key
        parsed_url = parse.urlparse(url)
        key = parse.urlunparse(
            parse.ParseResult._make(
                [parsed_url.scheme, parsed_url.netloc, '', '', '', '']
            )
        )
        session.auth = self.session_auth.get(key)
        return session

    def _chunks(self, response: 'requests.Response') -> Iterator[bytes]:
        """Call progress hook for every chunk."""
        for data in super()._chunks(response=response):
            self._progress(bytes_new=len(data))
            yield data
